from os import path as osp

from marlben.config import Planting
from marlben.envs import GatheringConfig, GatheringConfigScripted
from marlben.scripted.environments.gathering import GatheringPlantingAgent



PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)


class PlantingGatheringConfig(GatheringConfig, Planting):
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = f'maps/planting_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'



class PlantingGatheringConfigScripted(GatheringConfigScripted, Planting):
    AGENT_TYPE = GatheringPlantingAgent
    PLANTING_COST = 1 #
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")
        



from marlben.io.action.common import Fixed, NodeType, Node
from marlben.lib import material
from marlben.lib.utils import staticproperty
from marlben.lib.material import BalancedForest


class Plant(Node):
    priority = 4
    nodeType = NodeType.SELECTION

    @staticproperty
    def edges():
        return [PlantDecision]

    @staticproperty
    def leaf():
        return True

    def call(env, entity, plant_decision):
        if not env.config.game_system_enabled('Planting'):
            return

        if plant_decision == 0:
            return

        r, c = entity.history.lastPos

        tile = env.map.tiles[r, c]
        if tile.occupied or tile.lava or tile.mat == material.Water or tile.mat == material.Forest or tile.impassible:
            return

        if entity.status.freeze > 0:
            return

        if entity.resources.food.val >= env.config.PLANTING_COST:
            config = env.map.tiles[r, c].config
            tile = env.map.tiles[r, c]
            tile.reset(BalancedForest, True, True, config)
            env.map.harvest(r, c)
            entity.resources.food.decrement(env.config.PLANTING_COST)


class PlantDecision(Node):
    argType = Fixed

    @classmethod
    def N(cls, config):
        return 2

    def args(stim, entity, config):
        return [0, 1]

    @staticproperty
    def arg_name():
        return "plant_decision"



import copy
import random

import numpy as np

import marlben
from marlben.lib import material
from marlben.scripted import move, attack
from marlben.scripted.baselines import Scripted
from marlben.io.action import Move, Direction, East, West, South, North
from marlben import scripting

class GatheringAgent(Scripted):
    name = 'GatheringAgent_'
    '''Collects Resourses in a simple Gathering Env'''

    def __init__(self, config, idx):
        super().__init__(config, idx)

        self._current_target = "Water"
        self._forest_pos = 13
        self._water_pos = 10
        self._direction = (-1, 0)


    def _switch_direction(self, direction):
        return tuple([-1 * d for d in direction])

    def _manage_direction(self, r, c):
        if self._current_target == "Forest" and r == self._forest_pos:
            self._direction = self._switch_direction(self._direction)
            self._current_target = "Water"
        elif self._current_target == "Water" and r == self._water_pos:
            self._direction = self._switch_direction(self._direction)
            self._current_target = "Forest"

    def gather_actions(self):
        agent = self.ob.agent
        Entity = marlben.Serialized.Entity

        r = marlben.scripting.Observation.attribute(agent, Entity.R)
        c = marlben.scripting.Observation.attribute(agent, Entity.C)
        self._manage_direction(r, c)
        direction = move.towards(self._direction)
        self.actions[marlben.action.Move] = {marlben.action.Direction: direction}
        return self.actions

    def attack_actions(self, npc_only=True):
        self.scan_agents(npc_only=npc_only)
        if self.closest is not None:
            self.target = self.closest
            self.targetID = self.closestID
            self.targetDist = self.closestDist
        self.style = marlben.action.Melee
        self.attack()
        return self.actions

    def __call__(self, obs):
        super().__call__(obs)
        return self.gather_actions()


class GatheringBuildingAgent(GatheringAgent):
    def __init__(self, config, idx):
        super().__init__(config, idx)

        self._building_finished = False
        self._building_target_num = 0
        self._building_targets = [13, 10]
        self._need_building = False
        self._sidestep = -(self.iden % 2 * 2 - 1)

    def build_actions(self):
        agent = self.ob.agent
        Entity = marlben.Serialized.Entity
        r = marlben.scripting.Observation.attribute(agent, Entity.R)
        building = self._need_building

        if self._building_target_num < len(self._building_targets):
            target = self._building_targets[self._building_target_num]
            direction = target - r
            if direction == 0:
                self._building_target_num += 1
                direction = (0, -self._sidestep)
                direction = move.towards(direction)
                self._sidestep *= -1
            else:
                direction = int(direction / abs(direction))
                direction = move.towards((direction, 0))

        self._need_building = self._building_target_num > 0

        self.actions[marlben.action.Move] = {marlben.action.Direction: direction}
        self.actions[marlben.action.Build] = {marlben.action.BuildDecision: building}

        self._building_finished = self._building_target_num > 1

        return self.actions

    def __call__(self, obs):
        super().__call__(obs)
        if not self._building_finished:
            return self.build_actions()
        return self.gather_actions()


class ObscuredAndExclusiveGatheringAgent(Scripted):
    def __init__(self, config, idx):
        super().__init__(config, idx)
        self.known_water_sources = set()
        self.known_food_sources = set()
        self.forbidden_water_sources = set()
        self.forbidden_food_sources = set()
        self.doubtful_food_sources = set()
        self.doubtful_water_sources = set()

        self.unexplored = set([
            (r + self.config.TERRAIN_BORDER, c + self.config.TERRAIN_BORDER)
            for r in range(self.config.MAP_HEIGHT)
            for c in range(self.config.MAP_WIDTH)
        ])
        self.unvisited = set([
            (r + self.config.TERRAIN_BORDER, c + self.config.TERRAIN_BORDER)
            for r in range(self.config.MAP_HEIGHT)
            for c in range(self.config.MAP_WIDTH)
        ])
        self.resource_cooldowns = {}

        self.exploration_target = None
        self.prev_adjacent_tiles = None
        self.water_prev = None
        self.food_prev = None

    def explore(self):
        if (self.exploration_target is None or
            random.random() < ((len(self.unexplored) / (self.config.MAP_WIDTH * self.config.MAP_HEIGHT))**2) / 2 or
            abs(self.exploration_target[0] - self.currR) + abs(self.exploration_target[1] - self.currC) < self.config.NSTIM
        ):
            if len(self.unexplored) > 0:
                self.exploration_target = random.sample(self.unexplored, 1)[0]
            elif len(self.unvisited) > 0:
                self.exploration_target = random.sample(self.unvisited, 1)[0]

        move.pathfind(self.config, self.ob, self.actions,
                      self.exploration_target[0] - self.currR,
                      self.exploration_target[1] - self.currC)

    def __call__(self, obs):
        super().__call__(obs)

        for key in copy.deepcopy(set(self.resource_cooldowns.keys())):
            self.resource_cooldowns[key] -= 1
            if self.resource_cooldowns[key] <= 0:
                self.resource_cooldowns.pop(key)

        Tile = marlben.Serialized.Tile
        for tile in self.ob.tiles:
            coords = (round(marlben.scripting.Observation.attribute(tile, Tile.R)),
                      round(marlben.scripting.Observation.attribute(tile, Tile.C)))
            if coords in self.unexplored:
                self.unexplored.discard(coords)

            if marlben.scripting.Observation.attribute(tile, Tile.Index) != material.Grass.index:
                self.unvisited.discard(coords)

        coords = (self.currR, self.currC)
        self.unvisited.discard(coords)

        self._update_resources_register()
        resources_to_check = []

        if self.food < 3 * self.food_max / 4 and len(self.known_food_sources) > 0:
            # Go grab something to eat
            resources_to_check.append((self.food / self.food_max, self.known_food_sources))

        if self.water < 3 * self.water_max / 4 and len(self.known_water_sources) > 0:
            # Go get some water
            resources_to_check.append((self.water / self.water_max, self.known_water_sources))

        if len(resources_to_check) > 0:
            if len(resources_to_check) == 1 or resources_to_check[0][0] > resources_to_check[1][0]:
                self._move_to_resource(resources_to_check[0][1])
            else:
                self._move_to_resource(resources_to_check[1][1])
        else:
            # Explore for more resource sources
            self.explore()
        _steps = 0
        while not self._validate_movement_action() and _steps < 5:
            self.actions[Move][Direction] = np.random.choice([East, West, South, North])
            _steps += 1
        if not self._validate_movement_action():
            self.actions = {}
        return self.actions

    def _validate_movement_action(self):
        Tile = marlben.Serialized.Tile
        Observation = marlben.scripting.Observation
        if self.actions[Move][Direction] == North and Observation.attribute(self.ob.tile(-1, 0), Tile.Index) == material.Lava.index:
            return False
        if self.actions[Move][Direction] == South and Observation.attribute(self.ob.tile(+1, 0), Tile.Index) == material.Lava.index:
            return False
        if self.actions[Move][Direction] == East and Observation.attribute(self.ob.tile(0, +1), Tile.Index) == material.Lava.index:
            return False
        if self.actions[Move][Direction] == West and Observation.attribute(self.ob.tile(0, -1), Tile.Index) == material.Lava.index:
            return False
        return True

    def _update_resources_register(self):
        adjacent_tiles = [(self.currR+1, self.currC), (self.currR, self.currC+1),
                            (self.currR-1, self.currC), (self.currR, self.currC-1)]

        # If tile with resource is obscured
        if self.water_prev is not None and self.water_prev < self.water:
            near_source = False
            for t in self.prev_adjacent_tiles[0]:
                near_source = near_source or t in self.known_water_sources
            if not near_source:
                self.known_water_sources.add((self.prev_adjacent_tiles[1][0], self.prev_adjacent_tiles[1][1]))

        if self.food_prev is not None and self.food_prev < self.food:
            near_source = False
            for t in self.prev_adjacent_tiles[0]:
                near_source = near_source or t in self.known_food_sources
            if not near_source:
                self.known_food_sources.add((self.prev_adjacent_tiles[1][0], self.prev_adjacent_tiles[1][1]))

        self.prev_adjacent_tiles = (adjacent_tiles, (self.currR, self.currC))

        # If tile with resource is appeared to be private
        Tile = marlben.Serialized.Tile
        if self.food_prev is not None and self.water_prev is not None:
            for tile_pos in self.doubtful_food_sources:
                tile = self.ob.tile(tile_pos[0] - self.currR, tile_pos[1] - self.currC)
                dst = abs(tile_pos[0] - self.currR) + abs(tile_pos[1] - self.currC)
                tile_type = round(marlben.scripting.Observation.attribute(tile, Tile.Index))
                if tile_type in (material.Forest.index, material.BalancedForest.index) and self.food < self.food_prev and dst <= 1:
                    if tile_pos in self.known_food_sources:
                        self.known_food_sources.discard(tile_pos)
                    self.forbidden_food_sources.add(tile_pos)

            for tile_pos in self.doubtful_water_sources:
                tile = self.ob.tile(tile_pos[0] - self.currR, tile_pos[1] - self.currC)
                dst = abs(tile_pos[0] - self.currR) + abs(tile_pos[1] - self.currC)
                tile_type = round(marlben.scripting.Observation.attribute(tile, Tile.Index))
                if tile_type in (material.Water.index, material.BalancedWater.index) and self.water < self.water_prev and dst <= 1:
                    if tile_pos in self.known_water_sources:
                        self.known_water_sources.discard(tile_pos)
                    self.forbidden_water_sources.add(tile_pos)

            self.doubtful_food_sources.clear()
            self.doubtful_water_sources.clear()

            for tile_pos in adjacent_tiles:
                tile = self.ob.tile(tile_pos[0] - self.currR, tile_pos[1] - self.currC)
                tile_type = round(marlben.scripting.Observation.attribute(tile, Tile.Index))

                if tile_type in (material.Water.index, material.BalancedWater.index) and self.water < self.water_prev:
                    self.doubtful_water_sources.add(tile_pos)

                if tile_type in (material.Forest.index, material.BalancedForest.index) and self.food < self.food_prev:
                    self.doubtful_food_sources.add(tile_pos)

        # If see not obscured resource tile with unknown privacy
        for tile in self.ob.tiles:
            coords = (round(marlben.scripting.Observation.attribute(tile, Tile.R)),
                      round(marlben.scripting.Observation.attribute(tile, Tile.C)))
            tile_type = round(marlben.scripting.Observation.attribute(tile, Tile.Index))
            if tile_type in (material.Forest.index, material.BalancedForest.index) and coords not in self.forbidden_food_sources:
                self.known_food_sources.add(coords)
            if tile_type in (material.Water.index, material.BalancedWater.index) and coords not in self.forbidden_water_sources:
                self.known_water_sources.add(coords)

            if tile_type in (material.ScrubImpassible.index, material.Scrub.index):
                if coords not in self.resource_cooldowns:
                    self.resource_cooldowns[coords] = self.config.RESOURCE_COOLDOWN
                if coords not in self.known_water_sources and coords not in self.forbidden_water_sources and coords not in self.known_food_sources and coords not in self.forbidden_food_sources:
                    self.known_food_sources.add(coords)
                    self.known_water_sources.add(coords)

        self.water_prev = self.water
        self.food_prev = self.food

    def _move_to_resource(self, known_sources):
        any_target_source = None
        target_source = None
        target_distance = None
        for source in known_sources:
            if any_target_source is None:
                any_target_source = source
            if source in self.resource_cooldowns:
                continue
            distance = abs(source[0] - self.currR) + abs(source[1] - self.currC)
            if target_source is None or distance < target_distance:
                target_source = source
                target_distance = distance
        if target_source is None:
            target_source = any_target_source
        move.pathfind(self.config, self.ob, self.actions,
                      target_source[0] - self.currR, target_source[1] - self.currC)


class GatheringPlantingAgent(GatheringAgent):
    def __init__(self, config, idx):
        super().__init__(config, idx)

        self._planted = False
        self._forest_pos += 1

    def plant_actions(self):
        if self._planted:
            return self.actions
        agent = self.ob.agent
        Entity = marlben.Serialized.Entity
        r = marlben.scripting.Observation.attribute(agent, Entity.R)
        plant_decision = self._current_target == "Water" and r == self._forest_pos
        self._planted = plant_decision
        if self._planted:
            self._forest_pos -= 1
        self.actions[marlben.action.Plant] = {marlben.action.PlantDecision: plant_decision}
        return self.actions
    
    def __call__(self, obs):
        super().__call__(obs)
        self.gather_actions()
        return self.plant_actions()


class SiegeAgent(GatheringBuildingAgent):
    def __init__(self, config, idx):
        super().__init__(config, idx)

    def __call__(self, obs):
        super().__call__(obs)
        self.actions = self.attack_actions()
        return self.actions


class PveAgent(GatheringAgent):
    def __init__(self, config, idx):
        super().__init__(config, idx)

    def __call__(self, obs):
        super().__call__(obs)
        self.actions = self.attack_actions()
        return self.actions
        

class GatheringCombatAgent(GatheringAgent):
    def __init__(self, config, idx):
        super().__init__(config, idx)
        self._attacks = (self.iden + 1) % 2 == 0
        self._mid_col = 11 + int(not self._attacks)
    
    def fight_started(self):
        agent = self.ob.agent
        Entity = marlben.Serialized.Entity
        c = marlben.scripting.Observation.attribute(agent, Entity.C)
        return c == self._mid_col

    def go_to_fight_actions(self):
        agent = self.ob.agent
        Entity = marlben.Serialized.Entity
        c = marlben.scripting.Observation.attribute(agent, Entity.C)

        if c != self._mid_col:
            move_action = {Direction: East if self._attacks else West}
            self.actions[Move] = move_action
        return self.actions

    def __call__(self, obs):
        super().__call__(obs)
        if not self.fight_started():
            self.go_to_fight_actions()
        else:
            self.gather_actions()
        if self._attacks:
            self.attack_actions(npc_only=False)
        return self.actions



import torch
from torch import nn
from torch.nn.utils import rnn

from . import io
from .utils import nn_blocks


class Base(nn.Module):
    def __init__(self, config):
        '''Base class for baseline policies

        Args:
          config: A Configuration object
        '''
        super().__init__()
        self.embed = config.EMBED
        self.config = config

        self.input = io.Input(config,
                              embeddings=io.MixedEmbedding,
                              attributes=nn_blocks.SelfAttention)
        self.output = io.Output(config)

        self.valueF = nn.Linear(config.HIDDEN, 1)

    def hidden(self, obs, state=None, lens=None):
        '''Abstract method for hidden state processing, recurrent or otherwise,
        applied between the input and output modules

        Args:
          obs: An observation dictionary, provided by forward()
          state: The previous hidden state, only provided for recurrent nets
          lens: Trajectory segment lengths used to unflatten batched obs
        '''
        raise NotImplementedError('Implement this method in a subclass')

    def forward(self, obs, state=None, lens=None):
        '''Applies builtin IO and value function with user-defined hidden
        state subnetwork processing. Arguments are supplied by RLlib
        '''
        entityLookup = self.input(obs)
        hidden, state = self.hidden(entityLookup, state, lens)
        self.value = self.valueF(hidden).squeeze(1)
        actions = self.output(hidden, entityLookup)
        return actions, state


class Simple(Base):
    def __init__(self, config):
        '''Simple baseline model with flat subnetworks'''
        super().__init__(config)
        h = config.HIDDEN
        w = config.WINDOW()
        w_final = (w - 2) // 2
        self.ent = nn.Linear(2*h, h)
        self.conv = nn.Conv2d(h, h, 3)
        self.pool = nn.MaxPool2d(2)
        self.fc = nn.Linear(h * w_final * w_final, h)

        self.proj = nn.Linear(2*h, h)
        self.attend = nn_blocks.SelfAttention(self.embed, h)

    def hidden(self, obs, state=None, lens=None):
        # Attentional agent embedding
        agentEmb = obs['Entity']
        selfEmb = agentEmb[:, 0:1].expand_as(agentEmb)
        agents = torch.cat((selfEmb, agentEmb), dim=-1)
        agents = self.ent(agents)
        agents, _ = self.attend(agents)
        #agents = self.ent(selfEmb)

        # Convolutional tile embedding
        tiles = obs['Tile']
        self.attn = torch.norm(tiles, p=2, dim=-1)

        w = self.config.WINDOW()
        batch = tiles.size(0)
        hidden = tiles.size(2)
        # Dims correct?
        tiles = tiles.reshape(batch, w, w, hidden).permute(0, 3, 1, 2)
        tiles = self.conv(tiles)
        tiles = self.pool(tiles)
        tiles = tiles.reshape(batch, -1)
        tiles = self.fc(tiles)

        hidden = torch.cat((agents, tiles), dim=-1)
        hidden = self.proj(hidden)
        return hidden, state


class Recurrent(Simple):
    def __init__(self, config):
        '''Recurrent baseline model'''
        super().__init__(config)
        self.lstm = nn_blocks.BatchFirstLSTM(input_size=config.HIDDEN,
                                             hidden_size=config.HIDDEN)

    # Note: seemingly redundant transposes are required to convert between
    # Pytorch (seq_len, batch, hidden) <-> RLlib (batch, seq_len, hidden)
    def hidden(self, obs, state, lens):
        # Attentional input preprocessor and batching
        lens = lens.cpu() if type(lens) == torch.Tensor else lens
        hidden, _ = super().hidden(obs)
        config = self.config

        TB = hidden.size(0)  # Padded batch of size (seq x batch)
        B = len(lens)  # Sequence fragment time length
        TT = TB // B  # Trajectory batch size
        H = config.HIDDEN  # Hidden state size

        # Pack (batch x seq, hidden) -> (batch, seq, hidden)
        hidden = rnn.pack_padded_sequence(
            input=hidden.view(B, TT, H),
            lengths=lens,
            enforce_sorted=False,
            batch_first=True)

        # Main recurrent network
        hidden, state = self.lstm(hidden, state)

        # Unpack (batch, seq, hidden) -> (batch x seq, hidden)
        hidden, _ = rnn.pad_packed_sequence(sequence=hidden,
                                            batch_first=True,
                                            total_length=TT)

        return hidden.reshape(TB, H), state



import numpy as np

import torch
from torch import nn
from torch.nn import functional as F

from .utils import ModuleList


def Conv2d(fIn, fOut, k, stride=1):
    """torch Conv2d with same padding"""
    assert k % 2 == 0
    pad = int((k - 1) / 2)
    return torch.nn.Conv2d(fIn, fOut, k, stride=stride, padding=pad)


def Pool(k, stride=1, pad=0):
    return torch.nn.MaxPool2d(
        k, stride=stride, padding=pad)


class ConvReluPool(nn.Module):
    def __init__(self, fIn, fOut, k, stride=1, pool=2):
        super().__init__()
        self.conv = Conv2d(fIn, fOut, k, stride)
        self.pool = Pool(k)

    def forward(self, x):
        x = self.conv(x)
        x = F.relu(x)
        x = self.pool(x)
        return x


class ReluBlock(nn.Module):
    def __init__(self, h, layers=2, postRelu=True):
        super().__init__()
        self.postRelu = postRelu
        self.layers = ModuleList(nn.Linear, h, h, n=layers)

    def forward(self, x):
        for idx, fc in enumerate(self.layers):
            if idx != 0:
                x = torch.relu(x)
            x = fc(x)

        if self.postRelu:
            x = torch.relu(x)

        return x


class DotReluBlock(nn.Module):
    def __init__(self, h, layers=2):
        super().__init__()
        self.key = ReluBlock(h, layers, postRelu=False)
        self.val = ReluBlock(h, layers, postRelu=False)

    def forward(self, k, v):
        k = self.key(k).unsqueeze(-2)
        v = self.val(v)
        x = torch.sum(k * v, -1)
        return x


class ScaledDotProductAttention(nn.Module):
    def __init__(self, h):
        super().__init__()
        self.scale = np.sqrt(h)

    def forward(self, Q, K, V):
        Kt = K.transpose(-2, -1)
        QK = torch.matmul(Q, Kt)

        # Original is attending over hidden dims?
        # QK  = torch.softmax(QK / self.scale, dim=-2)
        QK = torch.softmax(QK / self.scale, dim=-1)
        score = torch.sum(QK, -2)
        QKV = torch.matmul(QK, V)
        return QKV, score


class SelfAttention(nn.Module):
    def __init__(self, xDim, yDim, flat=True):
        super().__init__()

        self.Q = torch.nn.Linear(xDim, yDim)
        self.K = torch.nn.Linear(xDim, yDim)
        self.V = torch.nn.Linear(xDim, yDim)

        self.attention = ScaledDotProductAttention(yDim)
        self.flat = flat

    def forward(self, q):
        Q = self.Q(q)
        K = self.K(q)
        V = self.V(q)

        attn, scores = self.attention(Q, K, V)

        if self.flat:
            attn, _ = torch.max(attn, dim=-2)

        return attn, scores


class BatchFirstLSTM(nn.LSTM):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, batch_first=True, **kwargs)

    def forward(self, input, hx):
        h, c = hx
        h = h.transpose(0, 1)
        c = c.transpose(0, 1)
        hidden, hx = super().forward(input, [h, c])
        h, c = hx
        h = h.transpose(0, 1)
        c = c.transpose(0, 1)
        return hidden, [h, c]



import numpy as np
from queue import PriorityQueue


def generate_huge_rocks(height, width, rock_seed_proba, rock_growth_proba,
                        max_growth_attempts=10, random: np.random.RandomState = None):
    prev_rocks = np.zeros((height, width))
    rocks = random.rand(height, width) < rock_seed_proba
    new_rocks = rocks
    attempts = 0
    while (prev_rocks != rocks).sum() > 0 and attempts < max_growth_attempts:
        candidates = np.logical_or(np.logical_or(np.roll(new_rocks, 1, 0), np.roll(new_rocks, -1, 0)),
                                   np.logical_or(np.roll(new_rocks, 1, 1), np.roll(new_rocks, -1, 1)))
        new_rocks = np.logical_and(candidates, random.rand(
            height, width) < rock_growth_proba)
        prev_rocks = rocks
        rocks = np.logical_or(rocks, new_rocks)
        attempts += 1

    return rocks


def generate_labyrinth(height, width, additional_links_min, additional_links_max, random: np.random.RandomState = None):
    assert height % 4 == 0 and width % 4 == 0
    h, w = height // 4, width // 4
    size = np.array([h, w])
    cells = np.zeros((h, w, 4))
    expandable_cells = PriorityQueue()
    expandable_cells.put(
        (random.random(), (random.randint(0, h), random.randint(0, w))))
    direction_helper = np.array([[1, 0], [0, 1], [0, -1], [-1, 0]])

    while expandable_cells.qsize() > 0:
        _, (x, y) = expandable_cells.get()
        directions = (size + direction_helper +
                      np.array([[x, y] for _ in range(4)])) % size
        available_directions = []
        for i, (tx, ty) in enumerate(directions):
            if cells[tx, ty].sum() == 0:
                available_directions.append(i)
        if len(available_directions) == 0:
            continue
        i = available_directions[random.randint(0, len(available_directions))]
        tx, ty = directions[i]
        cells[x, y, i] = 1
        cells[tx, ty, 3 - i] = 1
        expandable_cells.put((random.random(), (tx, ty)))
        if len(available_directions) > 1:
            expandable_cells.put((random.random(), (x, y)))

    for _ in range(random.randint(additional_links_min, additional_links_max + 1)):
        x, y, i = random.randint(0, h), random.randint(
            0, w), random.randint(0, 4)
        while cells[x, y, i] > 0:
            x, y = random.randint(0, h), random.randint(0, w),
            i = random.randint(0, 4)
        cells[x, y, i] = 1
        tx, ty = (size + direction_helper[i] + np.array([x, y])) % size
        cells[tx, ty, 3 - i] = 1

    map = np.zeros((height, width), dtype=bool)
    for x in range(h):
        for y in range(w):
            map[4 * x:4 * (x + 1), 4 * y:4 * (y + 1)] = True
            map[4 * x + 1:4 * (x + 1) - 1, 4 * y + 1:4 * (y + 1) - 1] = False
            if cells[x, y, 3] > 0:
                map[4 * x, 4 * y + 1:4 * (y + 1) - 1] = False
            if cells[x, y, 0] > 0:
                map[4 * (x + 1) - 1, 4 * y + 1:4 * (y + 1) - 1] = False
            if cells[x, y, 2] > 0:
                map[4 * x + 1:4 * (x + 1) - 1, 4 * y] = False
            if cells[x, y, 1] > 0:
                map[4 * x + 1:4 * (x + 1) - 1, 4 * (y + 1) - 1] = False
    map[0] = True
    map[-1] = True
    map[:, 0] = True
    map[:, -1] = True

    return map



import marlben

"""
In addition to agents, non-playable characters (NPCs) periodically appear on specific map cells (spawners) with deterministic behavior:
chase the nearest agent and attack him if possible.
Agents are granted an additional NPC attack action, the effect of which is to reduce the health of a random NPC in a small radius around the agent. 
That is, after attacking the NPC several times, it disappears from the map.

Expected behavior:
If all agents are focused on collecting resources and do not keep the number of NPCs under control during the episode, then at some point the number of NPCs can become too much.
Therefore, in order to stably survive for everyone, one of the agents needs to constantly cope with the NPC. In fact, timely control of NPCs is an element of the Public Good.
That is, all agents are better off if there are no NPCs on the map, but no agent is willing to risk their health and resources to secure this goal.
With this element, the environment resembles, for example, CleanUp, first introduced in Hughes et al. (2018).

Implementation details are located in config file

Hughes, E., Leibo, J.Z., Phillips, M., Tuyls, K., Duenez-Guzman, E.,
Castaneda, A.G., Dunning, I., Zhu, T., McKee, K., Koster, R. and Roff, H., 2018,
Inequity aversion improves cooperation in intertemporal social dilemmas. In
Proceedings of the 32nd International Conference on Neural Information
Processing Systems (pp. 3330-3340)..
"""
class Pve(marlben.Env):
    pass



import numpy as np

from marlben.io.action.common import NodeType, Node, Fixed
from marlben.lib.utils import staticproperty
from marlben.lib.distance import inRange


class Attack(Node):
    priority = 1
    nodeType = NodeType.SELECTION

    @staticproperty
    def n():
        return 4

    @staticproperty
    def edges():
        return [Style, Target]

    @staticproperty
    def leaf():
        return True

    def call(env, entity, style, targ):
        if entity.isPlayer and not env.config.game_system_enabled('Combat'):
            return

            # Check if self targeted
        if entity.entID == targ.entID:
            return

        stealing_enabled = env.config.STEALING_ENABLED

        # ADDED: POPULATION IMMUNITY
        # if entity.population == targ.population:
        #   return

        # Check attack range
        rng = style.attackRange(env.config)
        start = np.array(entity.base.pos)
        end = np.array(targ.base.pos)
        dif = np.max(np.abs(start - end))

        # Can't attack same cell or out of range
        if dif == 0 or dif > rng:
            return

            # Execute attack
        entity.history.attack = {}
        entity.history.attack['target'] = targ.entID
        entity.history.attack['style'] = style.__name__
        if style.__class__.__name__ != "Heal":
            # Healing is not considered an attack
            targ.attacker = entity
            targ.attackerID.update(entity.entID)

        from marlben.systems import combat
        dmg = combat.attack(entity, targ, style.skill, stealing_enabled)

        if style.freeze and dmg > 0:
            targ.status.freeze.update(env.config.COMBAT_FREEZE_TIME)

        return dmg


class Style(Node):
    argType = Fixed

    @staticproperty
    def edges():
        return [Melee, Range, Mage, Heal]

    def args(stim, entity, config):
        return Style.edges

    @staticproperty
    def arg_name():
        return "style"


class Target(Node):
    argType = None

    # argType = int

    @classmethod
    def N(cls, config):
        return config.N_AGENT_OBS

    def args(stim, entity, config):
        # Should pass max range?
        return inRange(entity, stim, config, None)

    @staticproperty
    def arg_name():
        return "targ"


class Melee(Node):
    nodeType = NodeType.ACTION
    index = 0
    freeze = False

    def attackRange(config):
        return config.COMBAT_MELEE_REACH

    def skill(entity):
        return entity.skills.melee


class Range(Node):
    nodeType = NodeType.ACTION
    index = 1
    freeze = False

    def attackRange(config):
        return config.COMBAT_RANGE_REACH

    def skill(entity):
        return entity.skills.range


class Mage(Node):
    nodeType = NodeType.ACTION
    index = 2
    freeze = True

    def attackRange(config):
        return config.COMBAT_MAGE_REACH

    def skill(entity):
        return entity.skills.mage


class Heal(Node):
    nodeType = NodeType.ACTION
    index = 3
    freeze = False

    def attackRange(config):
        return config.COMBAT_HEAL_REACH

    def skill(entity):
        return entity.skills.heal



class Observation:
    """Unwraps observation tensors for use with scripted agents"""

    def __init__(self, config, obs):
        """
        Args:
         config: A forge.blade.core.Config object or subclass object
         obs: An observation object from the environment
        """
        self.config = config
        self.obs = obs
        self.delta = config.NSTIM
        self.tiles = self.obs['Tile']['Continuous']
        self.agents = self.obs['Entity']['Continuous']
        self.n = int(self.obs['Entity']['N'])

    def tile(self, rDelta, cDelta):
        """Return the array object corresponding to a nearby tile

        Args:
         rDelta: row offset from current agent
         cDelta: col offset from current agent

        Returns:
         Vector corresponding to the specified tile
        """
        return self.tiles[self.config.WINDOW() * (self.delta + rDelta) + self.delta + cDelta]

    @property
    def agent(self):
        """Return the array object corresponding to the current agent"""
        return self.agents[0]

    @staticmethod
    def attribute(ary, attr):
        """
        Return an attribute of a game object

        Args:
         ary: The array corresponding to a game object
         attr: A forge.blade.io.stimulus.static stimulus class
        """
        return float(ary[attr.index])



import json
import os
from os import path as osp

import numpy as np
from tqdm import tqdm

from marlben import MapGenerator


class PregeneratedMapGenerator(MapGenerator):
    def __init__(self, config):
        super().__init__(config)

    def generate_all_maps(self):
        path_to_maps = osp.join(self.config.PATH_MAPS, 'maps.json')
        with open(path_to_maps, 'rb') as f:
            maps = json.load(f)
        height = self.config.MAP_HEIGHT
        width = self.config.MAP_WIDTH
        map_width = width + self.config.TERRAIN_BORDER * 2
        map_height = height + self.config.TERRAIN_BORDER * 2
        top, left = self.config.TERRAIN_BORDER, self.config.TERRAIN_BORDER
        for idx in tqdm(range(self.config.NMAPS)):
            map_template = np.zeros(
                shape=(map_height, map_width, 3), dtype=np.int32)
            tiles = np.array(maps[str(idx)], dtype=np.int32)
            map_template[top:top+height, left:left+width] = tiles
            # make rock border so the agent doesn't leave the main map
            map_template[[top-2, top+height+1], left-2:left+width+2, 0] = 5
            map_template[top-2:top+height+2, [left-2, left+width+1], 0] = 5
            path = osp.join(self.config.PATH_MAPS, 'map{}'.format(idx+1))
            os.makedirs(path, exist_ok=True)
            path = osp.join(path, "map.npy")
            np.save(path, map_template)



from os import path as osp
from marlben.config import NPC
from marlben.envs import BuildingGatheringConfig, BuildingGatheringConfigScripted
from marlben.envs.gathering.config import GatheringPlayerGroup

from marlben.core.spawn.spawn_system.position_samplers import UniformPositionSampler
from marlben.core.spawn.spawn_system.skill_samplers import CustomSkillSampler
from marlben.config.base.config import NPCGroupConfig, Config
from marlben.io.action.attack import Range, Heal, Mage, Melee

from marlben.scripted.environments.gathering import SiegeAgent



PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)

class MobsGroupConfig(NPCGroupConfig):
    BANNED_ATTACK_STYLES = {Heal, Range, Mage}
    DANGER = 1.0 
    SPAWN_ATTEMPTS_PER_ENT = 10

    def __init__(self, n_groups, agents_per_group, coord_sampler=None):
        super().__init__()
        if coord_sampler is not None:
            self.SPAWN_COORDINATES_SAMPLER = coord_sampler
        self.NENT = n_groups * agents_per_group
        self.SPAWN_SKILLS_SAMPLER = CustomSkillSampler(
            {"constitution": {"name": "const", "level": 2},
             "melee": {"name": "const", "level": 2}})

class SiegePlayerGroupConfig(GatheringPlayerGroup):
    BANNED_ATTACK_STYLES = [Range, Heal, Mage]

class SiegeConfig(BuildingGatheringConfig, NPC):
    PLAYER_GROUP_CONFIG = SiegePlayerGroupConfig
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.NPC_GROUPS = [MobsGroupConfig(n_groups, agents_per_group)]
        self.PATH_MAPS = f'maps/siege_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'


class SiegeConfigScripted(BuildingGatheringConfigScripted, NPC):
    PLAYER_GROUP_CONFIG = SiegePlayerGroupConfig
    AGENT_TYPE = SiegeAgent
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.NPC_GROUPS = [MobsGroupConfig(n_groups, agents_per_group,
            coord_sampler=UniformPositionSampler(r_range=[4, 4], c_range=[3, 5]))]
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")


import marlben
import numpy as np

import marlben.lib.distance
from . import utils


def closestTarget(config, ob, npc_only=False):
    shortestDist = np.inf
    closestAgent = None

    Entity = marlben.Serialized.Entity
    agent = ob.agent

    sr = marlben.scripting.Observation.attribute(agent, Entity.R)
    sc = marlben.scripting.Observation.attribute(agent, Entity.C)
    start = (sr, sc)

    for target in ob.agents:
        exists = marlben.scripting.Observation.attribute(target, Entity.Self)
        if not exists:
            continue

        if npc_only and marlben.scripting.Observation.attribute(target, Entity.ID) >= 0:
            continue

        tr = marlben.scripting.Observation.attribute(target, Entity.R)
        tc = marlben.scripting.Observation.attribute(target, Entity.C)

        goal = (tr, tc)
        dist = marlben.lib.distance.l1(start, goal)

        if dist < shortestDist and dist != 0:
            shortestDist = dist
            closestAgent = target

    if closestAgent is None:
        return None, None

    return closestAgent, shortestDist


def attacker(config, ob):
    Entity = marlben.Serialized.Entity

    sr = marlben.scripting.Observation.attribute(ob.agent, Entity.R)
    sc = marlben.scripting.Observation.attribute(ob.agent, Entity.C)

    attackerID = marlben.scripting.Observation.attribute(ob.agent, Entity.AttackerID)

    if attackerID == 0:
        return None, None

    for target in ob.agents:
        identity = marlben.scripting.Observation.attribute(target, Entity.ID)
        if identity == attackerID:
            tr = marlben.scripting.Observation.attribute(target, Entity.R)
            tc = marlben.scripting.Observation.attribute(target, Entity.C)
            dist = marlben.lib.distance.l1((sr, sc), (tr, tc))
            return target, dist
    return None, None


def target(config, actions, style, targetID):
    actions[marlben.action.Attack] = {
        marlben.action.Style: style,
        marlben.action.Target: targetID}



from marlben.envs import GatheringConfig


class ExclusiveGatheringConfig(GatheringConfig):

    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        super().__init__(n_groups, agents_per_group, tiles_per_agent, adjust_resource_amount)
        self.RESOURCE_COOLDOWN = 4
        self.RESOURCE_BASE_RESOURCE = 40
        self.RESOURCE_HARVEST_RESTORE_FRACTION = (self.RESOURCE_COOLDOWN / self.RESOURCE_BASE_RESOURCE)**0.5
        self.NUM_ACCESSIBILITY_COLORS = n_groups
        for i, group in enumerate(self.PLAYER_GROUPS):
            group.ACCESSIBLE_COLORS = [i+1]
        self.PATH_MAPS = f'maps/colors_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'



import marlben

"""
This environment is a more competitive version of the Arena environment. Agents have an additional attack action that has an effect equivalent to that action in Arena.
The specificity lies in the fact that the agent receives an additional reward every time the health of one of the opponents drops to zero in any way.

Expected behavior:
Depending on the priorities of the agent, he can focus on survival or on attacking opponents. 
With a full focus on the latter, the agent will not live long and will receive a small or negative reward in total.
A more profitable strategy is to provide itself with enough resources to regenerate health, and only then show aggression towards opponents.
In addition, some agents can completely focus on survival, receiving a reward for any outcome of the battle between other agents.

Implementation details are located in config file
"""
class Deathmatch(marlben.Env):
    pass



"""Infrastructure layer for representing agent observations

Maintains a synchronized + serialized representation of agent observations in
flat tensors. This allows for fast observation processing as a set of tensor
slices instead of a lengthy traversal over hundreds of game properties.

Synchronization bugs are notoriously difficult to track down: make sure
to follow the correct instantiation protocol, e.g. as used for defining
agent/tile observations, when adding new types observations to the code"""
import copy
from collections import defaultdict

import marlben
import numpy as np


class DataType:
    CONTINUOUS = np.float32
    DISCRETE = np.int32


class Index:
    """Lookup index of attribute names"""

    def __init__(self, prealloc):
        self.free = {idx for idx in range(1, prealloc)}
        self.index = {}
        self.back = {}

    def full(self):
        return len(self.free) == 0

    def remove(self, key):
        row = self.index[key]
        del self.index[key]
        del self.back[row]

        self.free.add(row)
        return row

    def update(self, key):
        if key in self.index:
            row = self.index[key]
        else:
            row = self.free.pop()
            self.index[key] = row
            self.back[row] = key

        return row

    def get(self, key):
        return self.index[key]

    def teg(self, row):
        return self.back[row]

    def expand(self, cur, nxt):
        self.free.update({idx for idx in range(cur, nxt)})


class ContinuousTable:
    '''Flat tensor representation for a set of continuous attributes'''

    def __init__(self, config, obj, prealloc, dtype=DataType.CONTINUOUS):
        self.config = config
        self.dtype = dtype
        self.cols = {}
        self.nCols = 0

        for (attribute,), attr in obj:
            self.initAttr(attribute, attr)

        self.data = self.initData(prealloc, self.nCols)

    def initAttr(self, key, attr):
        if attr.CONTINUOUS:
            self.cols[key] = self.nCols
            self.nCols += 1

    def initData(self, nRows, nCols):
        return np.zeros((nRows, nCols), dtype=self.dtype)

    def update(self, row, attr, val):
        col = self.cols[attr]
        self.data[row, col] = val

    def expand(self, cur, nxt):
        data = self.initData(nxt, self.nCols)
        data[:cur] = self.data

        self.data = data
        self.nRows = nxt

    def get(self, rows, pad=None):
        data = self.data[rows]
        data[rows == 0] = 0

        if pad is not None:
            data = np.pad(data, ((0, pad - len(data)), (0, 0)))

        return data


class DiscreteTable(ContinuousTable):
    '''Flat tensor representation for a set of discrete attributes'''

    def __init__(self, config, obj, prealloc, dtype=DataType.DISCRETE):
        self.discrete, self.cumsum = {}, 0
        super().__init__(config, obj, prealloc, dtype)

    def initAttr(self, key, attr):
        if not attr.DISCRETE:
            return

        self.cols[key] = self.nCols

        # Flat index
        attr = attr(None, None, 0, config=self.config)
        self.discrete[key] = self.cumsum

        self.cumsum += attr.max - attr.min + 1
        self.nCols += 1

    def update(self, row, attr, val):
        col = self.cols[attr]
        self.data[row, col] = val + self.discrete[attr]


class Grid:
    '''Flat representation of tile/agent positions'''

    def __init__(self, R, C):
        self.data = np.zeros((R, C), dtype=np.int32)

    def zero(self, pos):
        r, c = pos
        self.data[r, c] = 0

    def set(self, pos, val):
        r, c = pos
        self.data[r, c] = val

    def move(self, pos, nxt, row):
        self.zero(pos)
        self.set(nxt, row)

    def window(self, rStart, rEnd, cStart, cEnd):
        crop = self.data[rStart:rEnd, cStart:cEnd].ravel()
        return list(filter(lambda x: x != 0, crop))


class GridTables:
    '''Combines a Grid + Index + Continuous and Discrete tables

   Together, these data structures provide a robust and efficient
   flat tensor representation of an entire class of observations,
   such as agents or tiles'''

    def __init__(self, config, obj, pad, prealloc=1000, expansion=2):
        self.grid = Grid(config.TERRAIN_SIZE, config.TERRAIN_SIZE)
        self.continuous = ContinuousTable(config, obj, prealloc)
        self.discrete = DiscreteTable(config, obj, prealloc)
        self.index = Index(prealloc)

        self.nRows = prealloc
        self.expansion = expansion
        self.radius = config.NSTIM
        self.pad = pad

    def get(self, ent, radius=None, entity=False):
        if radius is None:
            radius = self.radius

        r, c = ent.pos
        cent = self.grid.data[r, c]
        assert cent != 0

        rows = self.grid.window(
            r - radius, r + radius + 1,
            c - radius, c + radius + 1)

        # Self entity first
        if entity:
            rows.remove(cent)
            rows.insert(0, cent)

        values = {'Continuous': self.continuous.get(rows, self.pad),
                  'Discrete': self.discrete.get(rows, self.pad)}

        if entity:
            ents = [self.index.teg(e) for e in rows]
            assert ents[0] == ent.entID
            return values, ents

        return values

    def update(self, obj, val):
        key, attr = obj.key, obj.attr
        if self.index.full():
            cur = self.nRows
            self.nRows = cur * self.expansion

            self.index.expand(cur, self.nRows)
            self.continuous.expand(cur, self.nRows)
            self.discrete.expand(cur, self.nRows)

        row = self.index.update(key)
        if obj.DISCRETE:
            self.discrete.update(row, attr, val - obj.min)
        if obj.CONTINUOUS:
            self.continuous.update(row, attr, val)

    def move(self, key, pos, nxt):
        row = self.index.get(key)
        self.grid.move(pos, nxt, row)

    def init(self, key, pos):
        row = self.index.get(key)
        self.grid.set(pos, row)

    def remove(self, key, pos):
        self.index.remove(key)
        self.grid.zero(pos)


class Dataframe:
    '''Infrastructure wrapper class'''

    def __init__(self, config):
        self.config, self.data = config, defaultdict(dict)
        for (objKey,), obj in marlben.Serialized:
            self.data[objKey] = GridTables(config, obj, pad=obj.N(config))

    def update(self, node, val):
        self.data[node.obj].update(node, val)

    def remove(self, obj, key, pos):
        self.data[obj.__name__].remove(key, pos)

    def init(self, obj, key, pos):
        self.data[obj.__name__].init(key, pos)

    def move(self, obj, key, pos, nxt):
        self.data[obj.__name__].move(key, pos, nxt)

    def get(self, ent):
        stim = {}

        stim['Entity'], ents = self.data['Entity'].get(ent, entity=True)
        stim['Entity']['N'] = np.array([len(ents)], dtype=np.int32)

        ent.targets = ents
        stim['Tile'] = self.__apply_visibility_to_tiles(
            ent, self.data['Tile'].get(ent))

        return stim

    def __apply_visibility_to_tiles(self, ent, tiles):
        accessibility_col = self.data["Tile"].discrete.cols["AccessibilityColor"]
        accessibility_offset = self.data["Tile"].discrete.discrete["AccessibilityColor"]
        visibility_col = self.data["Tile"].discrete.cols["VisibilityColor"]
        visibility_offset = self.data["Tile"].discrete.discrete["VisibilityColor"]
        visibility_c_col = self.data["Tile"].continuous.cols["VisibilityColor"]
        accessibility_c_col = self.data["Tile"].continuous.cols["AccessibilityColor"]
        index_col = self.data["Tile"].discrete.cols["Index"]
        index_c_col = self.data["Tile"].continuous.cols["Index"]

        new_tiles = {"Continuous": [], "Discrete": []}
        for (row_d, row_c) in zip(tiles["Discrete"], tiles["Continuous"]):
            new_row_d = copy.deepcopy(row_d)
            new_row_c = copy.deepcopy(row_c)
            if (row_d[visibility_col] - visibility_offset) not in ent.visible_colors:
                new_row_d[index_col] = 5
                new_row_c[index_c_col] = 5.
            new_row_d[visibility_col] = visibility_offset
            new_row_d[accessibility_col] = accessibility_offset
            new_row_c[visibility_c_col] = 0.
            new_row_c[accessibility_c_col] = 0.
            new_tiles["Continuous"].append(new_row_c)
            new_tiles["Discrete"].append(new_row_d)
        new_tiles["Continuous"] = np.array(new_tiles["Continuous"])
        new_tiles["Discrete"] = np.array(new_tiles["Discrete"])
        return new_tiles



from pdb import set_trace as T
import numpy as np

import random

import marlben
from marlben.entity import entity
from marlben.systems import combat, equipment, ai, combat, skill
from marlben.lib.colors import Neon

"""
Represents NPC on the map with a specified spawn place and set of actions
"""
class NPC(entity.Entity):
    def __init__(self, realm, pos, iden, name, color, pop, skills):
        super().__init__(realm, pos, iden, name, color, pop, skills)

    def update(self, realm, actions):
        super().update(realm, actions)

        if not self.alive:
            return

        self.resources.health.increment(1)
        self.lastAction = actions

    @staticmethod
    def spawn(realm, pos, iden, skills, custom_danger=None):
        config = realm.config

        # Select AI Policy
        danger = combat.danger(config, pos) if custom_danger is None else custom_danger
        if danger >= config.NPC_SPAWN_AGGRESSIVE:
            ent = Aggressive(realm, pos, iden, skills)
        elif danger >= config.NPC_SPAWN_NEUTRAL:
            ent = PassiveAggressive(realm, pos, iden, skills)
        elif danger >= config.NPC_SPAWN_PASSIVE:
            ent = Passive(realm, pos, iden, skills)
        else:
            return

        # Set equipment levels
        ent.loadout.chestplate.level = NPC.gearLevel(skills.defense.level)
        ent.loadout.platelegs.level = NPC.gearLevel(skills.defense.level)

        return ent

    def yieldDrops(self):
        self.lastAttacker.receiveDrops(self.drops.roll())

    @staticmethod
    def gearLevel(lvl, offset=10):
        proposed = random.gauss(lvl-offset, offset)
        lvl = np.clip(proposed, 0, lvl)
        return int(lvl)

    @staticmethod
    def clippedLevels(config, danger, n=1):
        lmin = config.NPC_LEVEL_MIN
        lmax = config.NPC_LEVEL_MAX

        lbase = danger*(lmax-lmin) + lmin
        lspread = config.NPC_LEVEL_SPREAD

        lvlMin = int(max(lmin, lbase - lspread))
        lvlMax = int(min(lmax, lbase + lspread))

        lvls = [random.randint(lvlMin, lvlMax) for _ in range(n)]

        if n == 1:
            return lvls[0]

        return lvls

    def packet(self):
        data = super().packet()

        data['base'] = self.base.packet()
        data['skills'] = self.skills.packet()
        data['resource'] = {'health': self.resources.health.packet()}

        return data

    @property
    def isNPC(self) -> bool:
        return True


class Passive(NPC):
    def __init__(self, realm, pos, iden, skills):
        super().__init__(realm, pos, iden, 'Passive', Neon.GREEN, -1, skills)
        self.dataframe.init(marlben.Serialized.Entity, iden, pos)

    def decide(self, realm):
        return ai.policy.passive(realm, self)


class PassiveAggressive(NPC):
    def __init__(self, realm, pos, iden, skills):
        super().__init__(realm, pos, iden, 'Neutral', Neon.ORANGE, -2, skills)
        self.dataframe.init(marlben.Serialized.Entity, iden, pos)

    def decide(self, realm):
        return ai.policy.neutral(realm, self)


class Aggressive(NPC):
    def __init__(self, realm, pos, iden, skills):
        super().__init__(realm, pos, iden, 'Hostile', Neon.RED, -3, skills)
        self.dataframe.init(marlben.Serialized.Entity, iden, pos)
        self.vision = int(
            max(self.vision, 1 + combat.level(self.skills) // 10))
        self.dataframe.init(marlben.Serialized.Entity, self.entID, self.pos)

    def decide(self, realm):
        return ai.policy.hostile(realm, self)



import marlben
from os import path as osp
from .config import BossFightConfig

"""
A two-agent environment where the goal is to eliminate one strong NPC. 
Alone, no single agent can deal with NPCs, so achieving the goal requires coordination from the agents. 
A separate difficulty lies in the fact that the agents need to take damage in turn for mutual survival. 
Otherwise, the agent who takes all the damage will run out of health points earlier than the NPC.

Implementation details are located in config file
"""
class BossFight(marlben.Env):
    def __init__(self, config=None):
        if config is None:
            config = BossFightConfig()
        config.PATH_MAPS = osp.join(osp.abspath(osp.dirname(__file__)), "../../envs/boss_fight/maps")
        super().__init__(config)

    def step(self, actions):
        obs, rewards, dones, infos = super().step(actions)
        if self._done():
            for k in dones:
                dones[k] = True
        return obs, rewards, dones, infos

    def _done(self):
        return len(self.realm.entity_group_manager.npc_groups[0].dead) > 0



import numpy as np
import abc
from .base_samplers import UniformSampler, RangeSampler, ListSampler, ConstantSampler, ChoiceSampler
from marlben.systems.skill import SkillsBalanced
from marlben.systems import combat
import random


class SkillsSampler:
    def __init__(self):
        self.config = None

    @abc.abstractmethod
    def get_next(self, pos):
        pass

    def reset(self, config):
        self.config = config


class DefaultSkillSampler(SkillsSampler):
    def get_next(self, pos):
        return SkillsBalanced(self.config)


class CustomSkillSampler(SkillsSampler):
    def __init__(self, skill2config: dict):
        super().__init__()
        self.skill2sampler = dict(
            [(k, self._sampler_by_config(skill2config[k])) for k in skill2config])

    def _sampler_by_config(self, skill_sampler_config):
        if skill_sampler_config["name"] == "const":
            return ConstantSampler(skill_sampler_config["level"])
        elif skill_sampler_config["name"] == "choice":
            return ChoiceSampler(skill_sampler_config["level"])
        elif skill_sampler_config["name"] == "list":
            return ListSampler(skill_sampler_config["level"])

    def get_next(self, pos):
        skills = SkillsBalanced(self.config)
        for k in self.skill2sampler:
            skills.__getattribute__(k).setExpByLevel(
                self.skill2sampler[k].get_next())
        return skills

    def reset(self, config):
        super().reset(config)
        for s in self.skill2sampler.values():
            s.reset()


class DefaultNPCSkillSampler(SkillsSampler):
    def __init__(self, level_min, level_max, level_spread):
        super().__init__()
        self.level_min = level_min
        self.level_max = level_max
        self.level_spread = level_spread

    def get_next(self, pos):
        danger = combat.danger(self.config, pos)

        lmin = self.level_min
        lmax = self.level_max

        lbase = danger*(lmax-lmin) + lmin
        lspread = self.level_spread

        lvlMin = int(max(lmin, lbase - lspread))
        lvlMax = int(min(lmax, lbase + lspread))

        lvls = [random.randint(lvlMin, lvlMax) for _ in range(5)]

        skills = SkillsBalanced(self.config)
        constitution, defense, melee, ranged, mage = lvls

        skills.constitution.setExpByLevel(constitution)
        skills.defense.setExpByLevel(defense)
        skills.melee.setExpByLevel(melee)
        skills.range.setExpByLevel(ranged)
        skills.mage.setExpByLevel(mage)

        return skills



from os import path as osp

from marlben import Task, Agent
from marlben.config import Combat
from marlben.config.base.config import PlayerGroupConfig
from marlben.envs.gathering.config import GatheringConfig, GatheringConfigScripted
from marlben.io.action import Range, Mage, Melee

from marlben.scripted.environments.gathering import GatheringCombatAgent



PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)

class PreyKilledTask:
    __name__ = 'hunt'
    def __call__(self, realm, entity):
        if not entity.alive or realm.entity_group_manager.player_groups[entity.population].banned_attack_styles is not None:
            return 0

        # Amount of Prey died at this turn
        return sum(
            [len(group.dead) for group in realm.entity_group_manager.player_groups if group.group_id != entity.population]
        )

class PreyGroupConfig(PlayerGroupConfig):
    def __init__(self, agents, n_ent):
        super().__init__()
        self.NENT = n_ent
        self.AGENTS = agents

    BASE_HEALTH = 5
    BANNED_ATTACK_STYLES = [Range, Mage, Melee]


class PredatorGroupConfig(PlayerGroupConfig):
    def __init__(self, agents, n_ent):
        super().__init__()
        self.NENT = n_ent
        self.AGENTS = agents

    BASE_HEALTH = 10

class PredatorPreyConfig(GatheringConfig, Combat):
    TASKS = [Task(PreyKilledTask(), 1, 10.)]
    PLAYER_GROUPS = [PreyGroupConfig([Agent], 10), PredatorGroupConfig([Agent], 3)]
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = f'maps/predator_prey_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'

class PredatorPreyConfigScripted(GatheringConfigScripted, Combat):
    AGENT_TYPE = GatheringCombatAgent
    TASKS = [Task(PreyKilledTask(), 1, 10.)]
    PLAYER_GROUPS = [PreyGroupConfig([Agent], 10), PredatorGroupConfig([Agent], 3)]
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")



from marlben.lib import utils


class SequentialLoader:
    '''config.AGENT_LOADER that spreads out agent populations'''

    def __init__(self, config):
        items = config.AGENTS
        for idx, itm in enumerate(items):
            itm.policyID = idx

        self.items = items
        self.idx = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx = (self.idx + 1) % len(self.items)
        return self.idx, self.items[self.idx]


class Template(metaclass=utils.StaticIterable):
    def __init__(self):
        self.data = {}
        cls = type(self)

        # Set defaults from static properties
        for k, v in cls:
            self.set(k, v)

    def override(self, **kwargs):
        for k, v in kwargs.items():
            err = 'CLI argument: {} is not a Config property'.format(k)
            assert hasattr(self, k), err
            self.set(k, v)

    def set(self, k, v):
        if type(v) is not property:
            try:
                setattr(self, k, v)
            except:
                print('Cannot set attribute: {} to {}'.format(k, v))
                quit()
        self.data[k] = v

    def print(self):
        keyLen = 0
        for k in self.data.keys():
            keyLen = max(keyLen, len(k))

        print('Configuration')
        for k, v in self.data.items():
            print('   {:{}s}: {}'.format(k, keyLen, v))



import random

import numpy as np
from marlben import Terrain
from marlben.lib import material


def _walk_and_paint(important_cells, mask, reached_cells):
    new_passed_cells = np.logical_and(reached_cells, mask)
    new_reached_cells = reached_cells
    while new_reached_cells.sum() > 0:

        new_reached_cells = np.logical_or(
            np.logical_or(np.roll(new_passed_cells, 1, 0),
                          np.roll(new_passed_cells, -1, 0)),
            np.logical_or(np.roll(new_passed_cells, 1, 1),
                          np.roll(new_passed_cells, -1, 1))
        )

        new_passed_cells = np.logical_and(new_reached_cells, mask)
        new_passed_cells = np.logical_and(new_passed_cells, np.logical_not(reached_cells))

        reached_cells[:] = np.logical_or(reached_cells, new_reached_cells)

    return np.logical_and(important_cells, np.logical_not(reached_cells)).sum() == 0


def check_connectivity(map, start_points, connectivity_type):
    important_cells = np.zeros_like(map[:, :, 0])
    for mat in material.Resource:
        important_cells = np.logical_or(important_cells, map[:, :, 0] == mat.index)

    passable_tiles = np.ones_like(map[:, :, 0])
    for mat in material.Impassible.materials:
        passable_tiles = np.logical_and(passable_tiles, map[:, :, 0] != mat.index)

    if connectivity_type == "FULL":
        reached_cells = np.zeros_like(important_cells)
        pt = random.choice(start_points)
        reached_cells[pt[0], pt[1]] = True
        if not _walk_and_paint(important_cells, passable_tiles, reached_cells):
            return False
        return np.min(reached_cells[passable_tiles])
    elif connectivity_type == "STRONG":
        total_reached_cells = np.zeros_like(important_cells, dtype=int)
        for pt in start_points:
            reached_cells = np.zeros_like(important_cells)
            reached_cells[pt[0], pt[1]] = 1
            if not _walk_and_paint(important_cells, passable_tiles, reached_cells):
                return False
            total_reached_cells += reached_cells
        if np.min(total_reached_cells[passable_tiles]) == 0:
            return False
        return True
    elif connectivity_type == "WEAK":
        reached_cells = np.zeros_like(important_cells)
        for pt in start_points:
            reached_cells[pt[0], pt[1]] = 1
        return _walk_and_paint(important_cells, passable_tiles, reached_cells)

    return True



from marlben.config import Planting
from marlben.envs.corridor import ScriptedCorridorConfig, OneNeuralCorridorConfig, CorridorConfig, Corridor
from marlben.envs.boss_fight import BossFightConfig, BossFight
from marlben.envs.gathering import GatheringConfig, GatheringConfigScripted, Gathering
from marlben.envs.building import BuildingGatheringConfig, BuildingGatheringConfigScripted, GatheringBuilding
from marlben.envs.planting import PlantingGatheringConfig, PlantingGatheringConfigScripted, GatheringPlanting
from marlben.envs.raid import Raid
from marlben.envs.spying import ObscuredGatheringConfig, GatheringObscured
from marlben.envs.colors import ExclusiveGatheringConfig, GatheringExclusive
from marlben.envs.exploring import ObscuredAndExclusiveGatheringConfig, GatheringObscuredAndExclusive
from marlben.envs.siege import SiegeConfig, SiegeConfigScripted, Siege
from marlben.envs.pve import PveConfig, PveConfigScripted, Pve
from marlben.envs.deathmatch import DeathmatchConfig, DeathmatchConfigScripted, Deathmatch
from marlben.envs.team_deathmatch import TeamDeathmatchConfig, TeamDeathmatchConfigScripted, TeamDeathmatch
from marlben.envs.arena import ArenaConfig, ArenaConfigScripted, Arena
from marlben.envs.predator_prey import PredatorPreyConfig, PredatorPreyConfigScripted, PredatorPrey
from gymnasium.envs.registration import register


name2config = {
    "CorridorScripted": ScriptedCorridorConfig,
    "CorridorSinglePlayer": OneNeuralCorridorConfig,
    "Corridor": CorridorConfig,
    "Gathering": GatheringConfig,
    "GatheringScripted": GatheringConfigScripted,
    "GatheringBuilding": BuildingGatheringConfig,
    "GatheringBuildingScripted": BuildingGatheringConfigScripted,
    "GatheringPlanting": PlantingGatheringConfig,
    "GatheringPlantingScripted": PlantingGatheringConfigScripted,
    "GatheringObscured": ObscuredGatheringConfig,
    "GatheringExclusive": ExclusiveGatheringConfig,
    "GatheringObscuredAndExclusive": ObscuredAndExclusiveGatheringConfig,
    "Siege": SiegeConfig,
    "SiegeScripted": SiegeConfigScripted,
    "Pve": PveConfig,
    "PveScripted": PveConfigScripted,
    "Deathmatch": Deathmatch,
    "DeathmatchScripted": DeathmatchConfigScripted,
    "TeamDeathmatch": TeamDeathmatch,
    "TeamDeathmatchConfigScripted": TeamDeathmatchConfigScripted,
    "Arena": ArenaConfig,
    "ArenaScripted": ArenaConfigScripted,
    "PredatorPrey": PredatorPreyConfig,
    "PredatorPreyScripted": PredatorPreyConfigScripted
}

name2env = {
    #Easy
    'Gathering': Gathering,
    'Colors': GatheringExclusive,
    'Pve': Pve,
    'BossFight': BossFight,
    'Building': GatheringBuilding,
    #Medium
    'Corridor': Corridor,
    'Planting': Planting,
    'Spying': GatheringObscured,
    'Raid': Raid,
    'Siege': Siege,
    #Hard
    'Arena': Arena,
    'Deathmatch': Deathmatch,
    'TeamDeathmatch': TeamDeathmatch,
    'Exploring': GatheringObscuredAndExclusive,
    'PredatorPrey': PredatorPrey
}

for key in name2env.keys():
    register(
        id='MARLBEN-' + key + '-v1',
        entry_point=name2env[key],
        disable_env_checker=True
    )





import os

import numpy as np
import vec_noise
from imageio import imread
from matplotlib.image import imsave
from scipy import stats as stats
from tqdm import tqdm

from marlben.lib import material


class MapGenerator:
    """Procedural map generation"""

    def __init__(self, config):
        self.config = config
        self.loadTextures()

    def loadTextures(self):
        """Called during setup; loads and resizes tile pngs"""
        lookup = {}
        path = self.config.PATH_TILE
        scale = self.config.MAP_PREVIEW_DOWNSCALE
        for mat in material.All:
            key = mat.tex
            tex = imread(path.format(key))
            lookup[mat.index] = tex[:, :, :3][::scale, ::scale]
            setattr(Terrain, key.upper(), mat.index)
        self.textures = lookup

    def generate_all_maps(self):
        '''Generates NMAPS maps according to generate_map

      Provides additional utilities for saving to .npy and rendering png previews'''

        config = self.config

        # Only generate if maps are not cached
        path_maps = os.path.join(config.PATH_CWD, config.PATH_MAPS)
        os.makedirs(path_maps, exist_ok=True)
        if not config.FORCE_MAP_GENERATION and os.listdir(path_maps):
            return

        if __debug__:
            print('Generating {} maps'.format(config.NMAPS))

        for idx in tqdm(range(config.NMAPS)):
            path = path_maps + '/map' + str(idx + 1)
            os.makedirs(path, exist_ok=True)

            terrain, tiles = self.generate_map(idx)

            # Save/render
            Save.np(tiles, path)
            if config.GENERATE_MAP_PREVIEWS:
                b = config.TERRAIN_BORDER
                tiles = [e[b:-b + 1] for e in tiles][b:-b + 1]
                Save.fractal(terrain, path + '/fractal.png')
                Save.render(tiles, self.textures, path + '/map.png')

    def generate_map(self, idx):
        '''Generate a single map

      The default method is a relatively complex multiscale perlin noise method.
      This is not just standard multioctave noise -- we are seeding multioctave noise
      itself with perlin noise to create localized deviations in scale, plus additional
      biasing to decrease terrain frequency towards the center of the map

      We found that this creates more visually interesting terrain and more deviation in
      required planning horizon across different parts of the map. This is by no means a
      gold-standard: you are free to override this method and create customized terrain
      generation more suitable for your application. Simply pass MAP_GENERATOR=YourMapGenClass
      as a config argument.'''

        config = self.config
        center = config.TERRAIN_CENTER
        border = config.TERRAIN_BORDER
        size = config.TERRAIN_SIZE
        frequency = config.TERRAIN_FREQUENCY
        offset = config.TERRAIN_FREQUENCY_OFFSET
        octaves = center // config.TERRAIN_TILES_PER_OCTAVE

        # Compute a unique seed based on map index
        # Flip seed used to ensure train/eval maps are different
        seed = idx + 8888
        if config.TERRAIN_FLIP_SEED:
            seed = -seed

        # Log interpolation factor
        if not hasattr(self, 'interpolaters'):
            self.interpolaters = np.logspace(config.TERRAIN_LOG_INTERPOLATE_MIN,
                                             config.TERRAIN_LOG_INTERPOLATE_MAX, config.NMAPS)

        interpolate = self.interpolaters[idx]

        # Data buffers
        val = np.zeros((size, size, octaves))
        scale = np.zeros((size, size, octaves))
        s = np.arange(size)
        X, Y = np.meshgrid(s, s)

        # Compute noise over logscaled octaves
        start = frequency
        end = min(start, start - np.log2(center) + offset)
        for idx, freq in enumerate(np.logspace(start, end, octaves, base=2)):
            val[:, :, idx] = vec_noise.snoise2(
                seed * size + freq * X, idx * size + freq * Y)

        # Compute L1 distance
        x = np.abs(np.arange(size) - size // 2)
        X, Y = np.meshgrid(x, x)
        data = np.stack((X, Y), -1)
        l1 = np.max(abs(data), -1)

        # Interpolation Weights
        rrange = np.linspace(-1, 1, 2 * octaves - 1)
        pdf = stats.norm.pdf(rrange, 0, interpolate)
        pdf = pdf / max(pdf)
        high = center / 2
        delta = high / octaves

        # Compute perlin mask
        noise = np.zeros((size, size))
        X, Y = np.meshgrid(s, s)
        expand = int(np.log2(center)) + 1
        for idx, octave in enumerate(range(expand, 1, -1)):
            freq, mag = 1 / 2 ** octave, 1 / 2 ** idx
            noise += mag * vec_noise.snoise2(seed * size + freq * X, idx * size + freq * Y)

        noise -= np.min(noise)
        noise = octaves * noise / np.max(noise) - 1e-12
        noise = noise.astype(int)

        # Compute L1 and Perlin scale factor
        for i in range(octaves):
            start = octaves - i - 1
            scale[l1 <= high] = np.arange(start, start + octaves)
            high -= delta

        start = noise - 1
        l1Scale = np.clip(l1, 0, size // 2 - border - 2)
        l1Scale = l1Scale / np.max(l1Scale)
        for i in range(octaves):
            idxs = l1Scale * scale[:, :, i] + (1 - l1Scale) * (start + i)
            scale[:, :, i] = pdf[idxs.astype(int)]

        # Blend octaves
        std = np.std(val)
        val = val / std
        val = scale * val
        val = np.sum(scale * val, -1)
        val = std * val / np.std(val)
        val = 0.5 + np.clip(val, -1, 1) / 2

        # Threshold to materials
        matl = np.zeros((size, size), dtype=object)
        for y in range(size):
            for x in range(size):
                v = val[y, x]
                if v <= config.TERRAIN_WATER:
                    mat = Terrain.WATER
                elif v <= config.TERRAIN_GRASS:
                    mat = Terrain.GRASS
                elif v <= config.TERRAIN_FOREST:
                    mat = Terrain.FOREST
                else:
                    mat = Terrain.STONE
                matl[y, x] = mat

        # Lava and grass border
        matl[l1 > size / 2 - border] = Terrain.LAVA
        matl[l1 == size // 2 - border] = Terrain.GRASS

        edge = l1 == size // 2 - border - 1
        stone = (matl == Terrain.STONE) | (matl == Terrain.WATER)
        matl[edge & stone] = Terrain.FOREST

        return val, matl


class Save:
    '''Save utility for map files'''

    def render(mats, lookup, path):
        '''Render tiles to png'''
        images = [[lookup[e] for e in l] for l in mats]
        image = np.vstack([np.hstack(e) for e in images])
        imsave(path, image)

    def fractal(terrain, path):
        '''Render raw noise fractal to png'''
        frac = (256 * terrain).astype(np.uint8)
        imsave(path, frac)

    def np(mats, path):
        '''Save map to .npy'''
        path = os.path.join(path, 'map.npy')
        np.save(path, mats.astype(int))


class Terrain:
    '''Terrain material class; populated at runtime'''
    pass



import marlben
from marlben.io.action.common import NodeType, Node, Fixed
from marlben.lib.utils import staticproperty


class Move(Node):
    priority = 2
    nodeType = NodeType.SELECTION

    def call(env, entity, direction):
        r, c = entity.pos
        entID = entity.entID
        entity.history.lastPos = (r, c)
        rDelta, cDelta = direction.delta
        rNew, cNew = r + rDelta, c + cDelta

        # One agent per cell
        tile = env.map.tiles[rNew, cNew]
        if tile.occupied and not tile.lava:
            return

        if entity.status.freeze > 0:
            return

        if tile.accessibility_color not in entity.accessible_colors:
            return

        env.dataframe.move(marlben.Serialized.Entity, entID, (r, c), (rNew, cNew))
        entity.base.r.update(rNew)
        entity.base.c.update(cNew)

        env.map.tiles[r, c].delEnt(entID)
        env.map.tiles[rNew, cNew].addEnt(entity)

        if env.map.tiles[rNew, cNew].lava:
            entity.receiveDamage(None, entity.resources.health.val)

    @staticproperty
    def edges():
        return [Direction]

    @staticproperty
    def leaf():
        return True


class Direction(Node):
    argType = Fixed

    @staticproperty
    def edges():
        return [North, South, East, West]

    def args(stim, entity, config):
        return Direction.edges

    @staticproperty
    def arg_name():
        return "direction"


class North(Node):
    index = 0
    delta = (-1, 0)


class South(Node):
    index = 1
    delta = (1, 0)


class East(Node):
    index = 2
    delta = (0, 1)


class West(Node):
    index = 3
    delta = (0, -1)



from collections import defaultdict

import numpy as np
from ray.rllib import MultiAgentEnv
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.algorithms.ppo import PPOConfig, PPO
from ray.rllib.algorithms.appo import APPOConfig, APPO
from ray.rllib.algorithms.impala import ImpalaConfig, Impala
from tqdm import tqdm

import marlben
from marlben.lib import log


class RLlibEnv(marlben.Env, MultiAgentEnv):
    """Wrapper class for using Neural MMO with RLlib"""

    def __init__(self, config):
        self.config = config["config"]
        super().__init__(self.config)

    def render(self):
        # Patch for RLlib dupe rendering bug
        if not self.config.RENDER:
            return

        super().render()

    def step(self, decisions):
        obs, rewards, dones, infos = super().step(decisions)
        config = self.config
        ts = config.TEAM_SPIRIT

        if config.COOPERATIVE:
            # Union of task rewards across population
            team_rewards = defaultdict(lambda: defaultdict(int))
            populations = {}
            for entID, info in infos.items():
                pop = info.pop('population')
                populations[entID] = pop
                team = team_rewards[pop]
                for task, reward in info.items():
                    team[task] = max(team[task], reward)

            # Team spirit interpolated between agent and team summed task rewards
            for entID, reward in rewards.items():
                pop = populations[entID]
                rewards[entID] = ts * sum(team_rewards[pop].values()) + (1 - ts) * reward

        dones['__all__'] = False

        if config.EVALUATE:
            horizon = config.EVALUATION_HORIZON
        else:
            horizon = config.TRAIN_HORIZON

        hit_horizon = self.realm.tick >= horizon
        truncated = {'__all__': False}

        if (hit_horizon or len(self.realm.players()) == 0):
            dones['__all__'] = True
            truncated['__all___'] = True

        return obs, rewards, dones, truncated, infos

    def reset(self, seed=None, options=None, idx=None, step=True):
        self.has_reset = True

        self.actions = {}
        self.dead = []

        self.quill = log.Quill()

        if idx is None:
            idx = np.random.randint(self.config.NMAPS) + 1

        self.worldIdx = idx
        self.realm.reset(idx)

        if step:
            self.obs, _, _, _, infos = self.step({})

        return self.obs, infos


class RLlibOverlayRegistry(marlben.OverlayRegistry):
    """Host class for RLlib Map overlays"""

    def __init__(self, realm):
        super().__init__(realm.config, realm)

        self.overlays['values'] = Values
        self.overlays['attention'] = Attention
        self.overlays['tileValues'] = TileValues
        self.overlays['entityValues'] = EntityValues


class RLlibOverlay(marlben.Overlay):
    '''RLlib Map overlay wrapper'''

    def __init__(self, config, realm, trainer, model):
        super().__init__(config, realm)
        self.trainer = trainer
        self.model = model


class Attention(RLlibOverlay):
    def register(self, obs):
        '''Computes local attentional maps with respect to each agent'''
        tiles = self.realm.realm.map.tiles
        players = self.realm.realm.players

        attentions = defaultdict(list)
        for idx, playerID in enumerate(obs):
            if playerID not in players:
                continue
            player = players[playerID]
            r, c = player.pos

            rad = self.config.NSTIM
            obTiles = self.realm.realm.map.tiles[r - rad:r + rad + 1, c - rad:c + rad + 1].ravel()

            for tile, a in zip(obTiles, self.model.attention()[idx]):
                attentions[tile].append(float(a))

        sz = self.config.TERRAIN_SIZE
        data = np.zeros((sz, sz))
        for r, tList in enumerate(tiles):
            for c, tile in enumerate(tList):
                if tile not in attentions:
                    continue
                data[r, c] = np.mean(attentions[tile])

        colorized = marlben.overlay.twoTone(data)
        self.realm.register(colorized)


class Values(RLlibOverlay):
    def update(self, obs):
        '''Computes a local value function by painting tiles as agents
      walk over them. This is fast and does not require additional
      network forward passes'''
        players = self.realm.realm.players
        for idx, playerID in enumerate(obs):
            if playerID not in players:
                continue
            r, c = players[playerID].base.pos
            self.values[r, c] = float(self.model.value_function()[idx])

    def register(self, obs):
        colorized = marlben.overlay.twoTone(self.values[:, :])
        self.realm.register(colorized)


def zeroOb(ob, key):
    for k in ob[key]:
        ob[key][k] *= 0


class GlobalValues(RLlibOverlay):
    '''Abstract base for global value functions'''

    def init(self, zeroKey):
        if self.trainer is None:
            return

        print('Computing value map...')
        model = self.trainer.get_policy('policy_0').model
        obs, ents = self.realm.dense()
        values = 0 * self.values

        # Compute actions to populate model value function
        BATCH_SIZE = 128
        batch = {}
        final = list(obs.keys())[-1]
        for agentID in tqdm(obs):
            ob = obs[agentID]
            batch[agentID] = ob
            zeroOb(ob, zeroKey)
            if len(batch) == BATCH_SIZE or agentID == final:
                self.trainer.compute_actions(batch, state={}, policy_id='policy_0')
                for idx, agentID in enumerate(batch):
                    r, c = ents[agentID].base.pos
                    values[r, c] = float(self.model.value_function()[idx])
                batch = {}

        print('Value map computed')
        self.colorized = marlben.overlay.twoTone(values)

    def register(self, obs):
        print('Computing Global Values. This requires one NN pass per tile')
        self.init()

        self.realm.register(self.colorized)


class TileValues(GlobalValues):
    def init(self, zeroKey='Entity'):
        '''Compute a global value function map excluding other agents. This
      requires a forward pass for every tile and will be slow on large maps'''
        super().init(zeroKey)


class EntityValues(GlobalValues):
    def init(self, zeroKey='Tile'):
        '''Compute a global value function map excluding tiles. This
      requires a forward pass for every tile and will be slow on large maps'''
        super().init(zeroKey)


class Trainer:
    def __init__(self, config, env=None, logger_creator=None):
        super().__init__(config, env, logger_creator)

    @classmethod
    def name(cls):
        return cls.__bases__[0].__name__

    def post_mean(self, stats):
        for key, vals in stats.items():
            if type(vals) == list:
                stats[key] = np.mean(vals)

    def train(self):
        stats = super().train()
        self.post_mean(stats['custom_metrics'])
        return stats

    def evaluate(self):
        return super().evaluate()


def PPOCustom(config):
    class PPOCustom(Trainer, PPO): pass

    ppoConfig = PPOConfig()
    ppoConfig.sgd_minibatch_size = config.SGD_MINIBATCH_SIZE
    return PPOCustom, ppoConfig


def APPOCustom(config):
    class APPOCustom(Trainer, APPO): pass

    return APPOCustom, APPOConfig()


def ImpalaCustom(config):
    class ImpalaCustom(Trainer, Impala): pass

    return ImpalaCustom, ImpalaConfig


###############################################################################
### Logging
class RLlibLogCallbacks(DefaultCallbacks):
    def on_episode_end(self, *, worker, base_env, policies, episode, **kwargs):
        assert len(base_env.envs) == 1, 'One env per worker'
        env = base_env.envs[0]

        stats = env.terminal()['Stats']
        policy_ids = stats.pop('PolicyID')

        for key, vals in stats.items():
            policy_stat = defaultdict(list)

            # Per-population metrics
            for policy_id, v in zip(policy_ids, vals):
                policy_stat[policy_id].append(v)

            for policy_id, vals in policy_stat.items():
                k = f'{policy_id}_{key}'
                episode.custom_metrics[k] = np.mean(vals)

        if not env.config.EVALUATE:
            return

        episode.custom_metrics['Raw_Policy_IDs'] = policy_ids
        episode.custom_metrics['Raw_Task_Rewards'] = stats['Task_Reward']



import marlben

"""
As the name suggests, this environment is the team version of the deathmatch environment. 
Agents are divided into multiple teams. The attack action has been modified so that its effect can only be applied to an agent from the other team.

Expected behavior: 
Naive agents can behave similarly to Arena or Deathmatch environments, ignoring team affiliation. 
A more profitable strategy is to coordinate with your team members to attack opponents, thus providing additional rewards and reducing competition for resources.

Implementation details are located in config file
"""
class TeamDeathmatch(marlben.Env):
    pass



from enum import Enum

from .base import spawn_concurrent, spawn_continuous, spawn_in_range


class SpawnFactory:
    class Type(Enum):
        CONCURRENT = 'concurrent'
        CONTINUOUS = 'continuous'
        IN_RANGE = 'in_range'

    def get_spawn_system(spawn_system_name):
        mode = SpawnFactory.Type(spawn_system_name)
        if mode == SpawnFactory.Type.CONCURRENT:
            return spawn_concurrent
        if mode == SpawnFactory.Type.CONTINUOUS:
            return spawn_continuous
        if mode == SpawnFactory.Type.IN_RANGE:
            return spawn_in_range



from ray.rllib.models.torch.recurrent_net import RecurrentNetwork
from torch import nn
import torch
from .policy import Recurrent


class RLlibPolicy(RecurrentNetwork, nn.Module):
    '''Wrapper class for using our baseline models with RLlib'''

    def __init__(self, *args, **kwargs):
        self.config = kwargs.pop('config')
        super().__init__(*args, **kwargs)
        nn.Module.__init__(self)

        # self.space  = actionSpace(self.config).spaces
        self.model = Recurrent(self.config)

    # Initial hidden state for RLlib Trainer
    def get_initial_state(self):
        return [self.model.valueF.weight.new(1, self.config.HIDDEN).zero_(),
                self.model.valueF.weight.new(1, self.config.HIDDEN).zero_()]

    def forward(self, input_dict, state, seq_lens):
        logitDict, state = self.model(input_dict['obs'], state, seq_lens)

        logits = []
        # Flatten structured logits for RLlib
        # TODO: better to use the space directly here in case of missing keys
        for atnKey, atn in sorted(logitDict.items()):
            for argKey, arg in sorted(atn.items()):
                logits.append(arg)

        return torch.cat(logits, dim=1), state

    def value_function(self):
        return self.model.value

    def attention(self):
        return self.model.attn



import ray

from marlben.config import Planting, Config
from marlben.envs import Corridor, BossFight, GatheringBuilding, GatheringExclusive, GatheringObscuredAndExclusive, \
    GatheringObscured
from marlben.envs.arena.env import Arena
from marlben.envs.deathmatch.env import Deathmatch
from marlben.envs.gathering.env import Gathering
from marlben.envs.predator_prey import PredatorPrey
from marlben.envs.pve.env import Pve
from marlben.envs.raid import Raid
from marlben.envs.siege.env import Siege
from marlben.envs.team_deathmatch.env import TeamDeathmatch
from .rllib_wrapper import RLlibEnv
from .neural import rllib_policy
from marlben.rllib.config import get_config

env_factory = {
    'Corridor': Corridor,
    'Gathering': Gathering,
    'Arena': Arena,
    'Pve': Pve,
    'Deathmatch': Deathmatch,
    'TeamDeathmatch': TeamDeathmatch,
    'BossFight': BossFight,
    'Building': GatheringBuilding,
    'Colors': GatheringExclusive,
    'Exploring': GatheringObscuredAndExclusive,
    'Spying': GatheringObscured,
    'Raid': Raid,
    'Siege': Siege,
    'Planting': Planting,
    'PredatorPrey': PredatorPrey
}


for key in env_factory.keys():
    class CustomEnv(env_factory[key], RLlibEnv):
        def __init__(self, config):
            super().__init__(config)

    ray.tune.registry.register_env(key, lambda config: CustomEnv(config))

ray.tune.registry.register_env('base', lambda config: RLlibEnv(config))



import numpy as np

from torch import nn


def modelSize(net):
    """Print model size"""
    params = 0
    for e in net.parameters():
        params += np.prod(e.size())
    params = int(params / 1000)
    print("Network has ", params, "K params")


def ModuleList(module, *args, n=1):
    """Repeat module n times"""
    return nn.ModuleList([module(*args) for i in range(n)])



import heapq
import itertools
import itertools


class PriorityQueue:
    def __init__(self, capacity, unique=False):
        self.q, self.items = [], set()
        self.capacity = capacity
        self.count = itertools.count()
        self.unique = unique

    def get(self, ind):
        priority, item = self.tolist()[ind]
        return priority, item

    def push(self, item, priority, uniqueKey=None):
        if self.unique:
            self.items.add(uniqueKey)
        count = next(self.count)
        if len(self.q) >= self.capacity:
            return heapq.heappushpop(self.q, (priority, count, item))
        heapq.heappush(self.q, (priority, count, item))

    def pop(self):
        priority, _, item = heapq.heappop(self.q)
        if self.unique:
            self.items.remove(item)
        return priority, item

    @property
    def peek(self):
        return self.peekPriority, self.peekValue

    @property
    def peekPriority(self):
        ret = heapq.nlargest(1, self.q)
        if len(ret) > 0:
            return ret[0][0]

    @property
    def peekValue(self):
        ret = heapq.nlargest(1, self.q)
        if len(ret) > 0:
            return ret[0][2]

    def tolist(self):
        q = heapq.nlargest(self.n, self.q)
        return [(e[0], e[2]) for e in q]

    def priorities(self):
        return sorted([e[0] for e in self.q], reverse=True)

    def print(self):
        q = heapq.nlargest(self.n, self.q)
        print([(e[0]) for e in q], end='')
        print()

    @property
    def n(self):
        return len(self.q)



from .core.map_generation.base import MapGenerator, Terrain
from .systems.achievement import Task
from .core.env import Env
from .core.agent import Agent
from .core import agent
from .io.action import Action
from .io.stimulus import Serialized
from .io import action
from .overlay import Overlay, OverlayRegistry
from .lib.rating import OpenSkillRating
from .lib import material
from . import scripting
from . import envs
from . import config
from .version import __version__

import os

__rllib__ = ['rllib']
__all__ = ['Env', 'scripting', 'agent', 'Agent', 'Serialized', 'action', 'Action', 'material',
           'Task', 'Overlay', 'OverlayRegistry', 'envs', 'config', 'rllib', 'scripted']



import numpy as np

from marlben.io.action.attack import Target, Style
from marlben.io.action.common import NodeType, Node, Fixed
from marlben.lib.utils import staticproperty


class Share(Node):
    priority = 0
    nodeType = NodeType.SELECTION

    @staticproperty
    def edges():
        return [Resource, Target, ResourceAmount]

    @staticproperty
    def leaf():
        return True

    def call(env, entity, resource, targ, amount):
        if not env.config.game_system_enabled('Sharing') or not env.config.game_system_enabled('Resource'):
            return

        # Check if self targeted
        if entity.entID == targ.entID:
            return

        # Check sharing range
        rng = env.config.SHARING_DISTANCE
        start = np.array(entity.base.pos)
        end = np.array(targ.base.pos)
        dif = np.max(np.abs(start - end))

        # Can't share same cell or out of range or negative amount
        if dif == 0 or dif > rng or amount < 0:
            return

        # Execute attack
        entity.history.share = {'target': targ.entID,
                                'resource': resource.__name__, 'amount': amount}

        from marlben.systems import sharing
        true_amount = sharing.share(entity, targ, resource.resource, amount)

        return true_amount


class ResourceAmount(Node):
    argType = int

    @classmethod
    def N(cls, config):
        return config.N_AGENT_OBS

    def args(stim, entity, config):
        # Should pass max range?
        return list(range(config.SHARE_MIN, config.SHARE_MAX + 1))

    @staticproperty
    def arg_name():
        return "amount"


class Resource(Node):
    argType = Fixed

    @staticproperty
    def edges():
        return [Water, Food]

    def args(stim, entity, config):
        return Style.edges

    @staticproperty
    def arg_name():
        return "resource"


class Water(Node):
    nodeType = NodeType.ACTION
    index = 0

    def resource(entity):
        return entity.resources.water


class Food(Node):
    nodeType = NodeType.ACTION
    index = 1

    def resource(entity):
        return entity.resources.food



from collections import defaultdict

import numpy as np
import torch
from torch import nn

import marlben
from .utils import nn_blocks


class MixedEmbedding(nn.Module):
    def __init__(self, continuous, discrete, config):
        super().__init__()

        self.continuous = torch.nn.ModuleList([torch.nn.Linear(1, config.EMBED) for _ in range(continuous)])
        self.discrete = torch.nn.Embedding(discrete, config.EMBED)

    def forward(self, x):
        continuous = x['Continuous'].split(1, dim=-1)
        continuous = [net(e) for net, e in zip(self.continuous, continuous)]
        continuous = torch.stack(continuous, dim=-2)
        discrete = self.discrete(x['Discrete'].long())

        return torch.cat((continuous, discrete), dim=-2)


class Input(nn.Module):
    def __init__(self, config, embeddings, attributes):
        '''Network responsible for processing observations

      Args:
         config     : A configuration object
         embeddings : An attribute embedding module
         attributes : An attribute attention module
      '''
        super().__init__()

        self.embeddings = nn.ModuleDict()
        self.attributes = nn.ModuleDict()

        for _, entity in marlben.Serialized:
            continuous = len([e for e in entity if e[1].CONTINUOUS])
            discrete = len([e for e in entity if e[1].DISCRETE])
            self.attributes[entity.__name__] = nn.Linear(
                (continuous + discrete) * config.HIDDEN, config.HIDDEN)
            self.embeddings[entity.__name__] = embeddings(
                continuous=continuous, discrete=4096, config=config)

        # TODO: implement obs scaling in a less hackey place
        self.tileWeight = torch.Tensor([1.0, 0.0, 0.02, 0.02, 1.0, 1.0])
        self.entWeight = torch.Tensor([1.0, 0.0, 0.0, 0.05, 0.00, 0.02, 0.02, 0.1, 0.01, 0.1, 0.1, 0.1, 0.3])

        if torch.cuda.is_available():
            self.tileWeight = self.tileWeight.cuda()
            self.entWeight = self.entWeight.cuda()

    def forward(self, inp):
        '''Produces tensor representations from an IO object 6082238515

      Args:                                                                   
         inp: An IO object specifying observations                      
         

      Returns:
         entityLookup      : A fixed size representation of each entity
      '''
        # Pack entities of each attribute set
        entityLookup = {}

        device = inp['Tile']['Continuous'].device
        inp['Tile']['Continuous'] *= self.tileWeight
        inp['Entity']['Continuous'] *= self.entWeight

        entityLookup['N'] = inp['Entity'].pop('N')
        for name, entities in inp.items():
            # Construct: Batch, ents, nattrs, hidden
            embeddings = self.embeddings[name](entities)
            B, N, _, _ = embeddings.shape
            embeddings = embeddings.view(B, N, -1)

            # Construct: Batch, ents, hidden
            entityLookup[name] = self.attributes[name](embeddings)

        return entityLookup


class Output(nn.Module):
    def __init__(self, config):
        '''Network responsible for selecting actions

      Args:
         config: A Config object
      '''
        super().__init__()
        self.config = config
        self.h = config.HIDDEN

        self.net = DiscreteAction(self.config, self.h, self.h)
        self.arg = nn.Embedding(marlben.Action.n, self.h)

    def names(self, nameMap, args):
        '''Lookup argument indices from name mapping'''
        return np.array([nameMap.get(e) for e in args])

    def forward(self, obs, lookup):
        '''Populates an IO object with actions in-place
                                                                              
      Args:                                                                   
         obs    : An IO object specifying observations
         lookup : A fixed size representation of each entity
      '''
        rets = defaultdict(dict)
        for atn in marlben.Action.edges:
            for arg in atn.edges:
                lens = None
                if arg.argType == marlben.action.Fixed:
                    batch = obs.shape[0]
                    if not 'edges' in arg.__dict__:
                        idxs = list(range(arg.N(self.config)))
                    else:
                        idxs = [e.idx for e in arg.edges]
                    cands = self.arg.weight[idxs]
                    cands = cands.repeat(batch, 1, 1)
                else:
                    cands = lookup['Entity']
                    lens = lookup['N']
                logits = self.net(obs, cands, lens)
                rets[atn][arg] = logits

        return rets


# Root action class
class Action(nn.Module):
    pass


class DiscreteAction(Action):
    '''Head for making a discrete selection from
   a variable number of candidate actions'''

    def __init__(self, config, xdim, h):
        super().__init__()
        self.net = nn_blocks.DotReluBlock(h)

    def forward(self, stim, args, lens):
        x = self.net(stim, args)
        if lens is not None:
            mask = torch.arange(x.shape[-1]).to(x.device).expand_as(x)
            x[mask >= lens] = 0

        return x



import random

from marlben.io.action import Move, Direction


class Agent:
    scripted = False
    name = 'Neural_'

    def __init__(self, config, idx):
        """Base class for agents

        Args:
           config: A Config object
           idx: Unique AgentID int
        """
        self.config = config
        self.iden = idx

    def __call__(self, obs):
        """Used by scripted agents to compute actions. Override in subclasses.

        Args:
            obs: Agent observation provided by the environment
        """
        pass


class Random(Agent):
    """Moves randomly, including bumping into things and falling into lava"""

    def __call__(self, obs):
        return {Move: {Direction: random.choice(Direction.edges)}}



import os
from copy import deepcopy

import ray
import torch
from ray import tune
from ray.tune import CLIReporter
from ray.tune.integration.wandb import WandbLoggerCallback

import marlben
from marlben.rllib import rllib_wrapper as wrapper


class ConsoleLog(CLIReporter):
    def report(self, trials, done, *sys_info):
        super().report(trials, done, *sys_info)


def setup_ray(config):
    torch.set_num_threads(1)
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['NUMEXPR_NUM_THREADS'] = '1'
    ray.init(local_mode=config.LOCAL_MODE)


def setup_policies(mapPolicy, config):
    policies = {}
    env = marlben.Env(config)
    for i in range(config.NPOLICIES):  # FIXME: Is it ok that we iterate through policies (not an agents)?
        params = {"agent_id": i,
                  "obs_space_dict": env.observation_space(i),
                  "act_space_dict": env.action_space(i)}
        key = mapPolicy(i, 0)
        policies[key] = (None, env.observation_space(i), env.action_space(i), params)
    return policies


def setup_wandb(callbacks):
    wandb_api_key = 'wandb_api_key'
    if os.path.exists(wandb_api_key):
        callbacks.append(WandbLoggerCallback(project='marlben', api_key_file='wandb_api_key', log_config=False))
    else:
        print('Running without WanDB. Create a file baselines/wandb_api_key and paste your API key to enable')


def get_restore(config, algorithm, config_name):
    restore = None
    if config.RESTORE:
        restore = '{0}/{1}/{2}/checkpoint_{3:06d}/checkpoint-{3}'.format(
            config.EXPERIMENT_DIR, algorithm, config_name, config.RESTORE_CHECKPOINT)
    return restore


def get_config_name(config):
    config_name = config.__class__.__name__
    if config.RESTORE and config.RESTORE_ID:
        config_name = '{}_{}'.format(config_name, config.RESTORE_ID)
    return config_name


def run_tune_experiment(config, env_name, trainer_wrapper):
    setup_ray(config)
    mapPolicy = lambda agentId, episode, worker=0: 'policy_{}'.format(agentId % config.NPOLICIES)
    policies = setup_policies(mapPolicy, config)

    # Evaluation config
    eval_config = deepcopy(config)
    eval_config.EVALUATE = True
    eval_config.AGENTS = []

    algo, base_config = trainer_wrapper(config)

    # Create rllib config
    rllib_config = base_config\
        .multi_agent(policies=policies,
                     policy_mapping_fn=mapPolicy,
                     count_steps_by='agent_steps') \
        .environment(env=env_name,
                     clip_actions=True,
                     env_config={"config": config},
                     disable_env_checking=True,
                     render_env=config.RENDER) \
        .evaluation(evaluation_config={"config": eval_config},
                    evaluation_interval=config.EVALUATION_INTERVAL,
                    evaluation_num_episodes=config.EVALUATION_NUM_EPISODES,
                    evaluation_num_workers=config.EVALUATION_NUM_WORKERS,
                    evaluation_parallel_to_training=config.EVALUATION_PARALLEL) \
        .rollouts(num_rollout_workers=config.NUM_WORKERS,
                  rollout_fragment_length=config.ROLLOUT_FRAGMENT_LENGTH,
                  num_envs_per_worker=1) \
        .callbacks(wrapper.RLlibLogCallbacks) \
        .training(train_batch_size=config.TRAIN_BATCH_SIZE,
                  lr=2e-5,
                  gamma=0.99,
                  lambda_=0.9,
                  use_gae=True,
                  clip_param=0.4,
                  grad_clip=None,
                  entropy_coeff=0.1,
                  vf_loss_coeff=0.25,
                  sgd_minibatch_size=64,
                  num_sgd_iter=config.NUM_SGD_ITER,
                  model={'custom_model': 'godsword',
                         'custom_model_config': {'config': config},
                         'max_seq_len': config.LSTM_BPTT_HORIZON}) \
        .debugging(log_level="ERROR") \
        .framework(framework="torch") \
        .resources(num_gpus=int(os.environ.get("RLLIB_NUM_GPUS", config.NUM_GPUS)),
                   num_gpus_per_worker=config.NUM_GPUS_PER_WORKER)

    callbacks = []
    setup_wandb(callbacks)
    config_name = get_config_name(config)

    tune.run(algo,
             config=rllib_config,
             verbose=config.LOG_LEVEL,
             stop={'training_iteration': config.TRAINING_ITERATIONS},
             restore=get_restore(config, algo.__name__, config_name),
             resume=config.RESUME,
             local_dir=config.EXPERIMENT_DIR,
             keep_checkpoints_num=config.KEEP_CHECKPOINTS_NUM,
             checkpoint_freq=config.CHECKPOINT_FREQ,
             checkpoint_at_end=True,
             trial_dirname_creator=lambda _: config_name,
             progress_reporter=ConsoleLog(),
             reuse_actors=True,
             callbacks=callbacks)



import random
import numpy as np


def spawn_concurrent(player_manager, config, realm):
    if player_manager.spawned:
        return

    player_manager.spawned = True
    idx = 0

    left = config.TERRAIN_BORDER
    right = config.TERRAIN_CENTER + config.TERRAIN_BORDER
    rrange = np.arange(left + 2, right, 4).tolist()

    assert not config.TERRAIN_CENTER % 4
    per_side = config.TERRAIN_CENTER // 4

    lows = (left + np.zeros(per_side, dtype=np.int)).tolist()
    highs = (right + np.zeros(per_side, dtype=np.int)).tolist()

    s1 = list(zip(rrange, lows))
    s2 = list(zip(lows, rrange))
    s3 = list(zip(rrange, highs))
    s4 = list(zip(highs, rrange))

    for r, c in s1 + s2 + s3 + s4:
        idx += 1
        assert not realm.map.tiles[r, c].occupied
        player_manager.spawnIndividual(r, c)


def spawn_continuous(player_manager, config, realm):
    # MMO-style spawning
    for _ in range(config.PLAYER_SPAWN_ATTEMPTS):
        if len(player_manager.entities) >= config.NENT:
            break
        mmax = config.TERRAIN_CENTER + config.TERRAIN_BORDER
        mmin = config.TERRAIN_BORDER
        var = np.random.randint(mmin, mmax)
        fixed = np.random.choice([mmin, mmax])
        r, c = int(var), int(fixed)
        if np.random.rand() > 0.5:
            r, c = c, r

        if realm.map.tiles[r, c].occupied:
            continue
        print('spawning individual')
        player_manager.spawnIndividual(r, c)

    while len(player_manager.entities) == 0:
        spawn_continuous(player_manager, config, realm)


def spawn_in_range(player_manager, config, realm):
    if player_manager.spawned:
        return
    player_manager.spawned = True
    top, left = config.TOP_LEFT_CORNER
    for _ in range(config.NENT):
        agent_idx = player_manager.idx - 1

        r_range = config.SPAWN_PARAMS['r_ranges'][agent_idx]
        r_range = [r + top for r in r_range]

        c_range = config.SPAWN_PARAMS['c_ranges'][agent_idx]
        c_range = [c + left for c in c_range]

        r = random.randint(*r_range)
        c = random.randint(*c_range)
        assert not realm.map.tiles[r, c].occupied
        player_manager.spawnIndividual(r, c)



from marlben.envs import name2config
from .presets.bases import Small
from .presets.scale import Debug
def get_config(env_name):
    EnvConfig = name2config[env_name]

    class Cfg(EnvConfig, Small, Debug):
        RENDER = False

    return Cfg


import marlben


"""
Agents are granted an additional attack action compared to Gathering env. Its effect is that the health of another random agent in a small radius is reduced.
This environment is suitable for testing the propensity for cooperation between agents.

Expected behavior:
On the one hand, the attack of other agents obviously helps to reduce their health to zero and, as a result, reduce the competition for resources.
On the other hand, agents can be guided by the principle of reciprocity, that is, respond to an attack with an attack, in which case it may be more profitable not to attack.
To avoid aggression from other agents, the agent can also simply avoid them.

Implementation details are located in config file
"""
class Arena(marlben.Env):
    pass



import inspect
from collections import deque

import numpy as np


class staticproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class Iterable(type):
    def __iter__(cls):
        queue = deque(cls.__dict__.items())
        while len(queue) > 0:
            name, attr = queue.popleft()
            if type(name) != tuple:
                name = tuple([name])
            if not inspect.isclass(attr):
                continue
            yield name, attr

    def values(cls):
        return [e[1] for e in cls]


class StaticIterable(type):
    def __iter__(cls):
        stack = list(cls.__dict__.items())
        for name, attr in cls.__dict__.items():
            if name == '__module__':
                continue
            if name.startswith('__'):
                break
            yield name, attr


class NameComparable(type):
    def __hash__(self):
        return hash(self.__name__)

    def __eq__(self, other):
        try:
            return self.__name__ == other.__name__
        except:
            print(
                'Some sphinx bug makes this block doc calls. You should not see this in normal NMMO usage')

    def __ne__(self, other):
        return self.__name__ != other.__name__

    def __lt__(self, other):
        return self.__name__ < other.__name__

    def __le__(self, other):
        return self.__name__ <= other.__name__

    def __gt__(self, other):
        return self.__name__ > other.__name__

    def __ge__(self, other):
        return self.__name__ >= other.__name__


class IterableNameComparable(Iterable, NameComparable):
    pass


def seed():
    return int(np.random.randint(0, 2**32))


def linf(pos1, pos2):
    r1, c1 = pos1
    r2, c2 = pos2
    return max(abs(r1 - r2), abs(c1 - c2))

# Bounds checker


def inBounds(r, c, shape, border=0):
    R, C = shape
    return (
        border < r < R - border and
        border < c < C - border
    )



import marlben

"""
As in the Building environment, agents are provided with an additional build action. Also, NPCs are implemented in the environment identical to the PVE environment.

Expected behavior:
In addition to using the build action to claim territory, agents can also fence off NPC spawners, thus making it impossible for NPCs to attack agents.
On the other hand, such a strategy requires resources and time, and therefore may not be beneficial to a particular agent in the absence of cooperation from other agents.

Implementation details are located in config file
"""
class Siege(marlben.Env):
    pass



import random as rand
from queue import PriorityQueue, Queue

import marlben
import numpy as np

import marlben.lib.distance
from marlben.lib import material

from . import utils


def adjacentPos(pos):
    r, c = pos
    return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1)]


def inSight(dr, dc, vision):
    return (
            -vision <= dr <= vision and
            -vision <= dc <= vision)


def vacant(tile):
    Tile = marlben.Serialized.Tile
    occupied = marlben.scripting.Observation.attribute(tile, Tile.NEnts)
    matl = marlben.scripting.Observation.attribute(tile, Tile.Index)

    return matl in material.Habitable and not occupied


def random(config, ob, actions):
    direction = rand.choice(marlben.action.Direction.edges)
    actions[marlben.action.Move] = {marlben.action.Direction: direction}


def towards(direction):
    if direction == (-1, 0):
        return marlben.action.North
    elif direction == (1, 0):
        return marlben.action.South
    elif direction == (0, -1):
        return marlben.action.West
    elif direction == (0, 1):
        return marlben.action.East
    else:
        return rand.choice(marlben.action.Direction.edges)


def pathfind(config, ob, actions, rr, cc):
    direction = aStar(config, ob, actions, rr, cc)
    direction = towards(direction)
    actions[marlben.action.Move] = {marlben.action.Direction: direction}


def meander(config, ob, actions):
    agent = ob.agent
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile

    r = marlben.scripting.Observation.attribute(agent, Entity.R)
    c = marlben.scripting.Observation.attribute(agent, Entity.C)

    cands = []
    if vacant(ob.tile(-1, 0)):
        cands.append((-1, 0))
    if vacant(ob.tile(1, 0)):
        cands.append((1, 0))
    if vacant(ob.tile(0, -1)):
        cands.append((0, -1))
    if vacant(ob.tile(0, 1)):
        cands.append((0, 1))
    if not cands:
        return (-1, 0)

    direction = rand.choices(cands)[0]
    direction = towards(direction)
    actions[marlben.action.Move] = {marlben.action.Direction: direction}


def explore(config, ob, actions, spawnR, spawnC):
    vision = config.NSTIM
    sz = config.TERRAIN_SIZE
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile

    agent = ob.agent
    r = marlben.scripting.Observation.attribute(agent, Entity.R)
    c = marlben.scripting.Observation.attribute(agent, Entity.C)

    centR, centC = sz // 2, sz // 2
    vR, vC = centR - spawnR, centC - spawnC

    mmag = max(abs(vR), abs(vC))
    rr = int(np.round(vision * vR / mmag))
    cc = int(np.round(vision * vC / mmag))
    pathfind(config, ob, actions, rr, cc)


def evade(config, ob, actions, attacker):
    Entity = marlben.Serialized.Entity

    sr = marlben.scripting.Observation.attribute(ob.agent, Entity.R)
    sc = marlben.scripting.Observation.attribute(ob.agent, Entity.C)

    gr = marlben.scripting.Observation.attribute(attacker, Entity.R)
    gc = marlben.scripting.Observation.attribute(attacker, Entity.C)

    rr, cc = (2 * sr - gr, 2 * sc - gc)

    pathfind(config, ob, actions, rr, cc)


def forageDijkstra(config, ob, actions, food_max, water_max, cutoff=100):
    vision = config.NSTIM
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile

    agent = ob.agent
    food = marlben.scripting.Observation.attribute(agent, Entity.Food)
    water = marlben.scripting.Observation.attribute(agent, Entity.Water)

    best = -1000
    start = (0, 0)
    goal = (0, 0)

    reward = {start: (food, water)}
    backtrace = {start: None}

    queue = Queue()
    queue.put(start)

    while not queue.empty():
        cutoff -= 1
        if cutoff <= 0:
            break

        cur = queue.get()
        for nxt in adjacentPos(cur):
            if nxt in backtrace:
                continue

            if not inSight(*nxt, vision):
                continue

            tile = ob.tile(*nxt)
            matl = marlben.scripting.Observation.attribute(tile, Tile.Index)
            occupied = marlben.scripting.Observation.attribute(tile, Tile.NEnts)

            if not vacant(tile):
                continue

            food, water = reward[cur]
            food = max(0, food - 1)
            water = max(0, water - 1)

            if matl == material.Forest.index:
                food = min(food + food_max // 2, food_max)
            for pos in adjacentPos(nxt):
                if not inSight(*pos, vision):
                    continue

                tile = ob.tile(*pos)
                matl = marlben.scripting.Observation.attribute(tile, Tile.Index)

                if matl == material.Water.index:
                    water = min(water + water_max // 2, water_max)
                    break

            reward[nxt] = (food, water)

            total = min(food, water)
            if total > best or (
                    total == best and max(food, water) > max(reward[goal])):
                best = total
                goal = nxt

            queue.put(nxt)
            backtrace[nxt] = cur

    while goal in backtrace and backtrace[goal] != start:
        goal = backtrace[goal]

    direction = towards(goal)
    actions[marlben.action.Move] = {marlben.action.Direction: direction}


def aStar(config, ob, actions, rr, cc, cutoff=100):
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile
    vision = config.NSTIM

    start = (0, 0)
    goal = (rr, cc)

    if start == goal:
        return (0, 0)

    pq = PriorityQueue()
    pq.put((0, start))

    backtrace = {}
    cost = {start: 0}

    closestPos = start
    closestHeuristic = marlben.lib.distance.l1(start, goal)
    closestCost = closestHeuristic

    while not pq.empty():
        # Use approximate solution if budget exhausted
        cutoff -= 1
        if cutoff <= 0:
            if goal not in backtrace:
                goal = closestPos
            break

        priority, cur = pq.get()

        if cur == goal:
            break

        for nxt in adjacentPos(cur):
            if not inSight(*nxt, vision):
                continue

            tile = ob.tile(*nxt)
            matl = marlben.scripting.Observation.attribute(tile, Tile.Index)
            occupied = marlben.scripting.Observation.attribute(tile, Tile.NEnts)

            # if not vacant(tile):
            #   continue

            if matl == material.Lava.index:
                continue  # We don't want to step on Lava, right?

            if occupied and not nxt == goal:
                continue

            # Omitted water from the original implementation. Seems key
            if matl in material.Impassible and not nxt == goal:
                continue

            newCost = cost[cur] + 1
            if nxt not in cost or newCost < cost[nxt]:
                cost[nxt] = newCost
                heuristic = utils.lInfty(goal, nxt)
                priority = newCost + heuristic

                # Compute approximate solution
                if heuristic < closestHeuristic or (
                        heuristic == closestHeuristic and priority < closestCost):
                    closestPos = nxt
                    closestHeuristic = heuristic
                    closestCost = priority

                pq.put((priority, nxt))
                backtrace[nxt] = cur

    # Not needed with scuffed material list above
    # if goal not in backtrace:
    #   goal = closestPos

    while goal in backtrace and backtrace[goal] != start:
        goal = backtrace[goal]

    return goal



def build_from_dict(config_dict):
    class CustomClass(*config_dict["presets_list"]):
        pass

    config = CustomClass()
    for key, value in config_dict["parameters"].items():
        setattr(config, key, value)

    return config



import marlben
from marlben.io.action import *
from marlben.lib import colors
from marlben.scripted import move
from marlben.scripted.baselines import Scripted


class BossFightTankAgent(Scripted):
    '''Simple Tank agent. Looks for a boss, then fights it until the end.'''
    name = 'TankAgent_'
    color = colors.Neon.GOLD
    scripted = True

    def __call__(self, obs):
        super().__call__(obs)

        Entity = marlben.Serialized.Entity

        self.scan_agents(True)
        if self.closestID is not None:
            self.style = Melee
            self.target = self.closest
            self.targetID = self.closestID
            self.targetDist = self.closestDist
            if self.closestDist > self.config.COMBAT_MELEE_REACH:
                move.pathfind(self.config, self.ob, self.actions,
                              marlben.scripting.Observation.attribute(self.closest, Entity.R) - self.currR,
                              marlben.scripting.Observation.attribute(self.closest, Entity.C) - self.currC)
            self.attack()
        else:
            self.explore()

        return self.actions


class BossRaidFighterAgent(Scripted):
    '''Simple Fighter agent. Looks for a boss, then fights it. Evades enemy if were attacked.'''
    name = 'FighterAgent_'
    color = colors.Neon.BLOOD
    is_evading = False

    def evade(self):
        move.evade(self.config, self.ob, self.actions, self.closest)

    def __call__(self, obs):
        super().__call__(obs)

        self.scan_agents(True)
        Entity = marlben.Serialized.Entity

        if self.closestID is not None:
            if self.closestDist <= self.config.COMBAT_MELEE_REACH:
                self.style = marlben.action.Melee
            else:
                self.style = marlben.action.Range
            self.target = self.closest
            self.targetID = self.closestID
            self.targetDist = self.closestDist
            self.attack()

            if self.attacker is None and not self.is_evading:
                if self.closestDist > self.config.COMBAT_MELEE_REACH:
                    move.pathfind(self.config, self.ob, self.actions,
                                  marlben.scripting.Observation.attribute(self.closest, Entity.R) - self.currR,
                                  marlben.scripting.Observation.attribute(self.closest, Entity.C) - self.currC)
            else:
                self.evade()
                # Evade until out of melee attack range
                if self.is_evading and self.closestDist > self.config.COMBAT_MELEE_REACH + 1:
                    self.is_evading = False
                else:
                    self.is_evading = True
        else:
            self.is_evading = False
            self.explore()

        return self.actions


class BossRaidHealerAgent(Scripted):
    '''Simple Healer agent. Follows an ally with the lowest HP remaining and heals it. Tries to evade enemy's attacks.'''
    name = 'HealerAgent_'
    color = colors.Neon.GREEN
    is_evading = False

    def evade(self):
        move.evade(self.config, self.ob, self.actions, self.closest)

    def __call__(self, obs):
        super().__call__(obs)

        self.scan_agents(True)
        self.allies = []
        Entity = marlben.Serialized.Entity
        for target in self.ob.agents:
            exists = marlben.scripting.Observation.attribute(target, Entity.Self)
            if not exists:
                continue
            if not marlben.scripting.Observation.attribute(target, Entity.ID) >= 0:
                continue
            self.allies.append(target)

        self.style = marlben.action.Heal

        self.lowestHealthID = None
        self.lowestHealth = None
        self.lowestHealthTarget = None
        for target in self.allies:
            target_pop = marlben.scripting.Observation.attribute(target, Entity.Population)
            curr_pop = marlben.scripting.Observation.attribute(self.ob.agent, Entity.Population)
            if target_pop == curr_pop:
                continue  # Do not follow and heal other healers

            # TODO: is there a way to get max entity health? It seems like its better to heal one with the lowest health percentage
            current_health = marlben.scripting.Observation.attribute(target, Entity.Health)
            if self.lowestHealthID is None or current_health < self.lowestHealth:
                self.lowestHealthID = marlben.scripting.Observation.attribute(target, Entity.ID)
                self.lowestHealth = current_health
                self.lowestHealthTarget = target

        self.actions = {}
        if self.lowestHealthID is not None:
            self.actions[Attack] = {Style: self.style, Target: self.lowestHealthID}
            if self.attacker is not None or (self.is_evading and self.closest is not None):
                self.evade()
                if self.is_evading and self.closestDist > self.config.COMBAT_MELEE_REACH + 1:
                    self.is_evading = False
                else:
                    self.is_evading = True
            else:
                move.pathfind(self.config, self.ob, self.actions,
                              marlben.scripting.Observation.attribute(self.lowestHealthTarget, Entity.R) - self.currR,
                              marlben.scripting.Observation.attribute(self.lowestHealthTarget, Entity.C) - self.currC)
        else:
            self.explore()

        return self.actions



from os import path as osp
from marlben.config import Combat
from marlben.envs.gathering.config import GatheringConfig, GatheringConfigScripted

from marlben.scripted.environments.gathering import GatheringCombatAgent



PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)


class ArenaConfig(GatheringConfig, Combat):
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = f'maps/arena_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'

class ArenaConfigScripted(GatheringConfigScripted, Combat):
    AGENT_TYPE = GatheringCombatAgent
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")


from collections import defaultdict
from typing import Dict
from marlben.core.spawn.base_manager import GroupsManager
from marlben import core, infrastructure


def prioritized(entities: Dict, merged: Dict):
    """Sort actions into merged according to priority"""
    for idx, actions in entities.items():
        for atn, args in actions.items():
            merged[atn.priority].append(
                (idx, (atn, dict([(k.arg_name, v) for k, v in args.items()]))))
    return merged


class Realm:
    """Top-level world object"""

    def __init__(self, config):
        self.config = config

        # Generate maps if they do not exist
        config.MAP_GENERATOR(config).generate_all_maps()

        # Load the world file
        self.dataframe = infrastructure.Dataframe(config)
        self.map = core.Map(config, self)

        # Entity handlers
        self.entity_group_manager = GroupsManager(config, self)

    def reset(self, idx):
        '''Reset the environment and load the specified map
      Args:
         idx: Map index to load
      '''
        print('----reset----')
        self.map.reset(self, idx)
        self.entity_group_manager.reset()
        self.tick = 0

    def packet(self):
        '''Client packet'''
        npc, player = self.entity_group_manager.packet
        return {'environment': self.map.repr,
                'resource': self.map.packet,
                'player': player,
                'npc': npc}

    @property
    def population(self):
        '''Number of player agents'''
        return self.entity_group_manager.players_count()

    def entity(self, entID):
        '''Get entity by ID'''
        return self.entity_group_manager.get_entity_by_id(entID)

    def players(self):
        players = []
        for player_group in self.entity_group_manager.player_groups:
            players.extend(player_group.entities.items())
        return players

    def agents(self):
        agents = []
        for player_group in self.entity_group_manager.player_groups:
            agents.extend(player_group.entities.keys())
        return agents

    def step(self, actions):
        '''Run game logic for one tick

        Args:
            actions: Dict of agent actions
        '''

        actions = self.entity_group_manager.mask_player_actions(actions)
        # Prioritize actions
        npcActions = self.entity_group_manager.get_npc_actions()
        merged = defaultdict(list)
        prioritized(actions, merged)
        prioritized(npcActions, merged)
        # Update entities and perform actions

        total_actions = dict()
        total_actions.update(actions)
        total_actions.update(npcActions)
        self.entity_group_manager.update(total_actions)

        # Execute actions
        for priority in sorted(merged):
            for entID, (atn, args) in merged[priority]:
                ent = self.entity(entID)
                atn.call(self, ent, **args)

        # Spawn new agent and cull dead ones
        # TODO: Place cull before spawn once PettingZoo API fixes respawn on same tick as death bug
        self.entity_group_manager.spawn()

        dead = self.entity_group_manager.cull()
        self.entity_group_manager.update_diary()

        # Update map
        self.map.step()
        self.tick += 1

        return dead



import numpy as np

from marlben.lib import utils


class SerializedVariable:
    CONTINUOUS = False
    DISCRETE = False

    def __init__(self, dataframe, key, val=None, config=None):
        if config is None:
            config = dataframe.config

        self.obj = str(self.__class__).split('.')[-2]
        self.attr = self.__class__.__name__
        self.key = key

        self.min = 0
        self.max = np.inf
        self.val = val

        self.dataframe = dataframe
        self.init(config)
        err = 'Must set a default val upon instantiation or init()'
        assert self.val is not None, err

        # Update dataframe
        if dataframe is not None:
            self.update(self.val)

    # Defined for cleaner stim files
    def init(self):
        pass

    def packet(self):
        return {
            'val': self.val,
            'max': self.max}

    def update(self, val):
        self.val = min(max(val, self.min), self.max)
        self.dataframe.update(self, self.val)
        return self

    def increment(self, val=1):
        self.update(self.val + val)
        return self

    def decrement(self, val=1):
        self.update(self.val - val)
        return self

    @property
    def empty(self):
        return self.val == 0

    def __add__(self, other):
        self.increment(other)
        return self

    def __sub__(self, other):
        self.decrement(other)
        return self

    def __eq__(self, other):
        return self.val == other

    def __ne__(self, other):
        return self.val != other

    def __lt__(self, other):
        return self.val < other

    def __le__(self, other):
        return self.val <= other

    def __gt__(self, other):
        return self.val > other

    def __ge__(self, other):
        return self.val >= other


class Continuous(SerializedVariable):
    CONTINUOUS = True


class Discrete(Continuous):
    DISCRETE = True


class Serialized(metaclass=utils.IterableNameComparable):

    class Entity(metaclass=utils.IterableNameComparable):
        @staticmethod
        def N(config):
            return config.N_AGENT_OBS

        class Self(Discrete):
            def init(self, config):
                self.max = 1
                self.scale = 1.0

        class ID(Continuous):
            def init(self, config):
                self.min = -np.inf
                self.scale = 0.001

        class AttackerID(Continuous):
            def init(self, config):
                self.min = -np.inf
                self.scale = 0.001

        class Level(Continuous):
            def init(self, config):
                self.scale = 0.05

        class Population(Discrete):
            def init(self, config):
                self.min = -3  # NPC index
                self.max = config.NPOP - 1
                self.scale = 1.0

        class R(Discrete):
            def init(self, config):
                self.min = 0
                self.max = config.TERRAIN_SIZE - 1
                self.scale = 0.15

        class C(Discrete):
            def init(self, config):
                self.min = 0
                self.max = config.TERRAIN_SIZE - 1
                self.scale = 0.15

        # Historical stats
        class Damage(Continuous):
            def init(self, config):
                # This scale may eventually be too high
                self.val = 0
                self.scale = 0.1

        class TimeAlive(Continuous):
            def init(self, config):
                self.val = 0
                self.scale = 0.01

        # Resources -- Redo the max/min scaling. You can't change these
        # after init without messing up the embeddings
        class Food(Continuous):
            def init(self, config):
                if config.game_system_enabled('Progression'):
                    self.val = config.PROGRESSION_BASE_RESOURCE
                    self.max = config.PROGRESSION_BASE_RESOURCE
                elif config.game_system_enabled('Resource'):
                    self.val = config.RESOURCE_BASE_RESOURCE
                    self.max = config.RESOURCE_BASE_RESOURCE
                else:
                    self.val = 1
                    self.max = 1

                self.scale = 0.1

        class Water(Continuous):
            def init(self, config):
                if config.game_system_enabled('Progression'):
                    self.val = config.PROGRESSION_BASE_RESOURCE
                    self.max = config.PROGRESSION_BASE_RESOURCE
                elif config.game_system_enabled('Resource'):
                    self.val = config.RESOURCE_BASE_RESOURCE
                    self.max = config.RESOURCE_BASE_RESOURCE
                else:
                    self.val = 1
                    self.max = 1

                self.scale = 0.1

        class Health(Continuous):
            def init(self, config):
                self.val = config.BASE_HEALTH
                self.max = config.BASE_HEALTH
                self.scale = 0.1

        # Status effects
        class Freeze(Continuous):
            def init(self, config):
                self.val = 0
                self.max = 3
                self.scale = 0.3

    class Tile(metaclass=utils.IterableNameComparable):
        @staticmethod
        def N(config):
            return config.WINDOW()**2

        class NEnts(Continuous):
            def init(self, config):
                self.max = config.NENT
                self.val = 0
                self.scale = 1.0

        class Index(Discrete):
            def init(self, config):
                self.max = config.NTILE
                self.scale = 0.15

        class R(Discrete):
            def init(self, config):
                self.max = config.TERRAIN_SIZE - 1
                self.scale = 0.15

        class C(Discrete):
            def init(self, config):
                self.max = config.TERRAIN_SIZE - 1
                self.scale = 0.15

        class AccessibilityColor(Discrete):
            def init(self, config):
                self.max = config.NUM_ACCESSIBILITY_COLORS
                self.val = 0
                self.scale = 1.

        class VisibilityColor(Discrete):
            def init(self, config):
                self.max = config.NUM_VISIBILITY_COLORS
                self.val = 0
                self.scale = 1.


for objName, obj in Serialized:
    for idx, (attrName, attr) in enumerate(obj):
        attr.index = idx



from pdb import set_trace as T
import numpy as np
from scipy import signal


def norm(ary, nStd=2):
    assert type(ary) == np.ndarray, 'ary must be of type np.ndarray'
    R, C = ary.shape
    preprocessed = np.zeros_like(ary)
    nonzero = ary[ary != 0]
    mean = np.mean(nonzero)
    std = np.std(nonzero)
    if std == 0:
        std = 1
    for r in range(R):
        for c in range(C):
            val = ary[r, c]
            if val != 0:
                val = (val - mean) / (nStd * std)
                val = np.clip(val+1, 0, 2)/2
                preprocessed[r, c] = val
    return preprocessed


def clip(ary):
    assert type(ary) == np.ndarray, 'ary must be of type np.ndarray'
    R, C = ary.shape
    preprocessed = np.zeros_like(ary)
    nonzero = ary[ary != 0]
    mmin = np.min(nonzero)
    mmag = np.max(nonzero) - mmin
    for r in range(R):
        for c in range(C):
            val = ary[r, c]
            val = (val - mmin) / mmag
            preprocessed[r, c] = val
    return preprocessed


def twoTone(ary, nStd=2, preprocess='norm', invert=False, periods=1):
    assert preprocess in 'norm clip none'.split()
    if preprocess == 'norm':
        ary = norm(ary, nStd)
    elif preprocess == 'clip':
        ary = clip(ary)

    R, C = ary.shape

    colorized = np.zeros((R, C, 3))
    if periods != 1:
        ary = np.abs(signal.sawtooth(periods*3.14159*ary))
    if invert:
        colorized[:, :, 0] = ary
        colorized[:, :, 1] = 1-ary
    else:
        colorized[:, :, 0] = 1-ary
        colorized[:, :, 1] = ary

    colorized *= (ary != 0)[:, :, None]

    return colorized



import numpy as np
from marlben import Terrain


def generate_resources(tile_map, resource_type, resource_amount, visibility_colors, accessibility_colors,
                       visibility_distribution, accessibility_distribution, random: np.random.RandomState = None):
    visibility_colors_counts = np.zeros(
        visibility_colors) if visibility_colors > 0 else None
    accessibility_colors_counts = np.zeros(
        accessibility_colors) if accessibility_colors > 0 else None
    height = tile_map.shape[0]
    width = tile_map.shape[1]

    for i in range(resource_amount):
        attempts = 0
        x, y = random.randint(0, height-1), random.randint(0, width-1)
        while tile_map[x, y, 0] != Terrain.GRASS and attempts < 20:
            x, y = random.randint(0, height-1), random.randint(0, width-1)
            attempts += 1
        assert attempts < 20
        tile_map[x, y, 0] = resource_type
        if visibility_colors > 0:
            tile_map[x, y, 1] = _get_random_color(
                visibility_colors_counts, visibility_distribution, random)
        if accessibility_colors > 0:
            tile_map[x, y, 2] = _get_random_color(
                accessibility_colors_counts, accessibility_distribution, random)


def _get_random_color(colors_counts, distribution, random: np.random.RandomState):
    if distribution == "UNIFORM":
        c = random.randint(1, len(colors_counts)+1)
    elif distribution == "EQUAL":
        args = np.argwhere(np.array(colors_counts) == np.min(colors_counts))
        c = args[np.random.randint(0, len(args))] + 1
    elif distribution == "FAIR":
        logits = np.max(colors_counts) - colors_counts + 1
        logits = logits / np.sum(logits)
        c = np.argmax(random.multinomial(1, logits)) + 1

    colors_counts[c-1] += 1
    return c



from pdb import set_trace as T
import abc

import numpy as np

from marlben.systems import experience, combat, ai

from marlben.lib import material

### Infrastructure ###


class SkillGroup:
    def __init__(self, config):
        self.expCalc = experience.ExperienceCalculator()
        self.config = config
        self.skills = set()

    def update(self, realm, entity, actions):
        for skill in self.skills:
            skill.update(realm, entity)

    def packet(self):
        data = {}
        for skill in self.skills:
            data[skill.name] = skill.packet()
        return data


class Skill:
    skillItems = abc.ABCMeta

    def __init__(self, skillGroup):
        self.config = skillGroup.config
        self.expCalc = skillGroup.expCalc
        self.exp = 0

        skillGroup.skills.add(self)

    @property
    def name(self):
        return self.__class__.__name__.lower()

    def packet(self):
        data = {}

        data['exp'] = self.exp
        data['level'] = self.level

        return data

    def update(self, realm, entity):
        pass

    def setExpByLevel(self, level):
        self.exp = self.expCalc.expAtLevel(level)

    @property
    def level(self):
        lvl = self.expCalc.levelAtExp(self.exp)
        assert lvl == int(lvl)
        return int(lvl)

### Skill Subsets ###


class Harvesting(SkillGroup):
    def __init__(self, config):
        super().__init__(config)

        self.fishing = Fishing(self)
        self.hunting = Hunting(self)


class HarvestingBalanced(SkillGroup):
    def __init__(self, config):
        super().__init__(config)

        self.fishing = CollectResource(self, "water")
        self.hunting = CollectResource(self, "food")


class Combat(SkillGroup):
    def __init__(self, config):
        super().__init__(config)

        self.constitution = Constitution(self)
        self.defense = Defense(self)
        self.melee = Melee(self)
        self.range = Range(self)
        self.mage = Mage(self)
        self.heal = Heal(self)

    def packet(self):
        data = super().packet()
        data['level'] = combat.level(self)

        return data

    def applyDamage(self, dmg, style):
        if not self.config.game_system_enabled('Progression'):
            return

        config = self.config
        baseScale = config.PROGRESSION_BASE_XP_SCALE
        combScale = config.PROGRESSION_COMBAT_XP_SCALE
        conScale = config.PROGRESSION_CONSTITUTION_XP_SCALE

        if dmg > 0.:
            self.constitution.exp += dmg * baseScale * conScale

        skill = self.__dict__[style]
        skill.exp += abs(dmg) * baseScale * combScale

    def receiveDamage(self, dmg, stealing_enabled=True):
        if not self.config.game_system_enabled('Progression') or dmg < 0.:
            return

        config = self.config
        baseScale = config.PROGRESSION_BASE_XP_SCALE
        combScale = config.PROGRESSION_COMBAT_XP_SCALE
        conScale = config.PROGRESSION_CONSTITUTION_XP_SCALE

        self.constitution.exp += dmg * baseScale * conScale
        self.defense.exp += dmg * baseScale * combScale


class Skills(Harvesting, Combat):
    pass


class SkillsBalanced(HarvestingBalanced, Combat):
    pass


### Individual Skills ###
class CombatSkill(Skill):
    pass


class Constitution(CombatSkill):
    def __init__(self, skillGroup):
        super().__init__(skillGroup)
        self.setExpByLevel(self.config.BASE_HEALTH)

    def update(self, realm, entity):
        health = entity.resources.health
        food = entity.resources.food
        water = entity.resources.water
        config = self.config

        if not config.game_system_enabled('Resource'):
            health.increment(1)
            return

        # Heal if above fractional resource threshold
        regen = config.RESOURCE_HEALTH_REGEN_THRESHOLD
        foodThresh = food > regen * entity.skills.hunting.level
        waterThresh = water > regen * entity.skills.fishing.level

        if foodThresh and waterThresh and config.REGEN_HEALTH:
            restore = config.RESOURCE_HEALTH_RESTORE_FRACTION
            restore = np.floor(restore * self.level)
            health.increment(restore)

        if food.empty:
            health.decrement(1)

        if water.empty:
            health.decrement(1)


class Melee(CombatSkill):
    pass


class Range(CombatSkill):
    pass


class Mage(CombatSkill):
    pass


class Heal(CombatSkill):
    pass


class Defense(CombatSkill):
    pass


class Fishing(Skill):
    def __init__(self, skillGroup):
        super().__init__(skillGroup)
        config, level = self.config, 1
        if config.game_system_enabled('Progression'):
            level = config.PROGRESSION_BASE_RESOURCE
        elif config.game_system_enabled('Resource'):
            level = config.RESOURCE_BASE_RESOURCE

        self.setExpByLevel(level)

    def update(self, realm, entity):
        if not self.config.game_system_enabled('Resource'):
            return

        water = entity.resources.water
        water.decrement(1)

        adj_tiles = ai.utils.adjacentTiles(realm.map.tiles, entity.pos)
        water_nearby = False
        for tile in adj_tiles:
            water_in_tile = (type(tile.state) in [material.Water] and
                             tile.accessibility_color in entity.accessible_colors)
            water_nearby = water_nearby or water_in_tile
        if not water_in_tile:
            return

        restore = self.config.RESOURCE_HARVEST_RESTORE_FRACTION
        restore = np.floor(restore * self.level)
        water.increment(restore)

        if self.config.game_system_enabled('Progression'):
            self.exp += self.config.PROGRESSION_BASE_XP_SCALE * restore


class Hunting(Skill):
    def __init__(self, skillGroup):
        super().__init__(skillGroup)
        config, level = self.config, 1
        if config.game_system_enabled('Progression'):
            level = config.PROGRESSION_BASE_RESOURCE
        elif config.game_system_enabled('Resource'):
            level = config.RESOURCE_BASE_RESOURCE

        self.setExpByLevel(level)

    def update(self, realm, entity):
        if not self.config.game_system_enabled('Resource'):
            return

        food = entity.resources.food
        food.decrement(1)

        r, c = entity.pos
        tile = realm.map.tiles[r, c]
        if (type(tile.mat) not in [material.Forest] or
                tile.accessibility_color not in entity.accessible_colors or
                not realm.map.harvest(r, c)):
            return

        restore = self.config.RESOURCE_HARVEST_RESTORE_FRACTION
        restore = np.floor(restore * self.level)
        food.increment(restore)

        if self.config.game_system_enabled('Progression'):
            self.exp += self.config.PROGRESSION_BASE_XP_SCALE * restore


class CollectResource(Skill):

    TYPE2MAT = {
        'food': material.BalancedForest,
        'water': material.BalancedWater
    }
    TYPE2NAME = {
        'food': 'hunting',
        'water': 'fishing'
    }

    def __init__(self, skillGroup, resource_type):
        super().__init__(skillGroup)
        config, level = self.config, 1
        if config.game_system_enabled('Progression'):
            level = config.PROGRESSION_BASE_RESOURCE
        elif config.game_system_enabled('Resource'):
            level = config.RESOURCE_BASE_RESOURCE

        self.setExpByLevel(level)
        self.resource_type = resource_type
        self.material_type = self.TYPE2MAT[resource_type]

    @property
    def name(self):
        return self.TYPE2NAME[self.resource_type]

    def update(self, realm, entity):
        if not self.config.game_system_enabled('Resource'):
            return
        resource = getattr(entity.resources, self.resource_type)
        resource.decrement(1)

        resource_positions = ai.utils.adjacentPosWithMat(
            realm.map.tiles, entity.pos, self.material_type
        )
        resource_positions = [(r, c) for r, c in resource_positions
                              if realm.map.tiles[r, c].accessibility_color in entity.accessible_colors]

        if not len(resource_positions):
            return

        r, c = resource_positions[0]
        realm.map.harvest(r, c)

        restore = self.config.RESOURCE_HARVEST_RESTORE_FRACTION
        restore = np.floor(restore * self.level)
        resource.increment(restore)

        if self.config.game_system_enabled('Progression'):
            self.exp += self.config.PROGRESSION_BASE_XP_SCALE * restore



from dataclasses import dataclass, field

import marlben


@dataclass
class RLlib:
    """Base config for RLlib Models

   Extends core Config, which contains environment, evaluation,
   and non-RLlib-specific learning parameters

   Configure NUM_GPUS and NUM_WORKERS for your hardware
   Note that EVALUATION_NUM_WORKERS cores are reserved for evaluation
   and one additional core is reserved for the driver process.
   Therefore, set NUM_WORKERS <= cores - EVALUATION_NUM_WORKERS - 1
    """

    # Run in train/evaluation mode
    EVALUATE: bool = False
    N_TRAIN_MAPS: int = 1

    @property
    def MODEL(self):
        return self.__class__.__name__

    @property
    def NMAPS(self):
        if not self.EVALUATE:
            return self.N_TRAIN_MAPS
        return super().NMAPS

    @property
    def TRAIN_BATCH_SIZE(self):
        if not hasattr(self, "_TRAIN_BATCH_SIZE") or self._TRAIN_BATCH_SIZE is None:
            self._TRAIN_BATCH_SIZE = 64 * 256 * self.NUM_WORKERS
        return self._TRAIN_BATCH_SIZE

    @TRAIN_BATCH_SIZE.setter
    def TRAIN_BATCH_SIZE(self, value):
        self._TRAIN_BATCH_SIZE = value

    # Checkpointing. Resume will load the latest trial, e.g. to continue training
    # Restore (overrides resume) will force load a specific checkpoint (e.g. for rendering)
    EXPERIMENT_DIR: str = 'experiments'
    RESUME: bool = False

    RESTORE: bool = False
    RESTORE_ID: str = 'Baseline'  # Experiment name suffix
    RESTORE_CHECKPOINT: int = 1000

    # Policy specification
    EVAL_AGENTS: list = field(default_factory=lambda: [marlben.Agent])
    AGENTS: list = field(default_factory=lambda: [marlben.Agent])
    TASKS: list = field(default_factory=lambda: [])

    # Hardware and debug
    NUM_GPUS_PER_WORKER: int = 0
    LOCAL_MODE: bool = False
    LOG_LEVEL: int = 1

    # Training and evaluation settings
    EVALUATION_INTERVAL: int = 1
    EVALUATION_PARALLEL: bool = False
    TRAINING_ITERATIONS: int = 1000
    KEEP_CHECKPOINTS_NUM: int = 3
    CHECKPOINT_FREQ: int = 1
    LSTM_BPTT_HORIZON: int = 16
    NUM_SGD_ITER: int = 1

    # Model
    SCRIPTED = None
    N_AGENT_OBS: int = 100
    NPOLICIES: int = 1
    HIDDEN: int = 64
    EMBED: int = 64

    # Reward
    COOPERATIVE: bool = False
    TEAM_SPIRIT: float = 0.0


class Small(RLlib, marlben.config.Small):
    '''Small scale Neural MMO training setting

   Features up to 64 concurrent agents and 32 concurrent NPCs,
   64 x 64 maps (excluding the border), and 128 timestep horizons'''

    # Memory/Batch Scale
    ROLLOUT_FRAGMENT_LENGTH: int = 128
    SGD_MINIBATCH_SIZE: int = 128

    # Horizon
    TRAIN_HORIZON: int = 128
    EVALUATION_HORIZON: int = 128


@dataclass
class Medium(RLlib, marlben.config.Medium):
    '''Medium scale Neural MMO training setting

   Features up to 256 concurrent agents and 128 concurrent NPCs,
   128 x 128 maps (excluding the border), and 1024 timestep horizons'''

    # Memory/Batch Scale
    ROLLOUT_FRAGMENT_LENGTH: int = 256
    SGD_MINIBATCH_SIZE: int = 128

    # Horizon
    TRAIN_HORIZON: int = 1024
    EVALUATION_HORIZON: int = 1024


@dataclass
class Large(RLlib, marlben.config.Large):
    """Large scale Neural MMO training setting

    Features up to 2048 concurrent agents and 1024 concurrent NPCs,
    1024 x 1024 maps (excluding the border), and 8192 timestep horizons
    """

    # Memory/Batch Scale
    ROLLOUT_FRAGMENT_LENGTH: int = 32
    SGD_MINIBATCH_SIZE: int = 128

    # Horizon
    TRAIN_HORIZON: int = 8192
    EVALUATION_HORIZON: int = 8192



from marlben.io.action.common import Fixed, NodeType, Node  # , BoolDecision
from marlben.lib.utils import staticproperty
from marlben.lib.material import Stone


class Build(Node):
    priority = 3
    nodeType = NodeType.SELECTION

    @staticproperty
    def edges():
        return [BuildDecision]

    @staticproperty
    def leaf():
        return True

    def call(env, entity, build_decision):
        if not env.config.game_system_enabled('Building'):
            return

        r, c = entity.history.lastPos

        tile = env.map.tiles[r, c]
        if tile.occupied or tile.lava:
            return

        if entity.status.freeze > 0:
            return

        if build_decision:
            config = env.map.tiles[r, c].config
            env.map.tiles[r, c].reset(Stone, True, True, config)


class BuildDecision(Node):
    argType = Fixed

    @classmethod
    def N(cls, config):
        return 2

    def args(stim, entity, config):
        return [0, 1]

    @staticproperty
    def arg_name():
        return "build_decision"
